from git.objects.commit import Commit

from jig.tests.testcase import JigTestCase
from jig.exc import GitRevListMissing, GitRevListFormatError
from jig.gitutils.branches import parse_rev_range


class TestParseRevRange(JigTestCase):

    """
    Git utils revision range parser.

    """
    def setUp(self):
        super(TestParseRevRange, self).setUp()

        self.gitrepo, self.gitrepodir, _ = self.repo_from_fixture('repo01')

    def assertIsRevRange(self, rev_range):
        commit_a, commit_b = rev_range

        self.assertIsInstance(commit_a, Commit)
        self.assertIsInstance(commit_b, Commit)

    def test_bad_format(self):
        """
        If the revision range doesn't match the expected format.
        """
        with self.assertRaises(GitRevListFormatError):
            parse_rev_range(self.gitrepodir, 'A-B')

    def test_bad_format_missing_rev(self):
        """
        If the format is correct but the revisions are missing.
        """
        with self.assertRaises(GitRevListFormatError):
            parse_rev_range(self.gitrepodir, '..B')

        with self.assertRaises(GitRevListFormatError):
            parse_rev_range(self.gitrepodir, 'A..')

    def test_bad_revs(self):
        """
        If the format is good but the revisions do not exist.
        """
        with self.assertRaises(GitRevListMissing):
            parse_rev_range(self.gitrepodir, 'FOO..BAR')

    def test_good_revs(self):
        """
        The revisions to exist.
        """
        self.assertIsRevRange(parse_rev_range(self.gitrepodir, 'HEAD^1..HEAD'))

    def test_local_branch(self):
        """
        A branch that is newly created can be referenced.
        """
        self.gitrepo.create_head('feature-branch')

        self.assertIsRevRange(
            parse_rev_range(self.gitrepodir, 'HEAD^1..feature-branch')
        )

    def test_out_of_range(self):
        """
        A revision is out of range.
        """
        with self.assertRaises(GitRevListMissing):
            parse_rev_range(self.gitrepodir, 'HEAD~1000..HEAD')



from os.path import expanduser

from jig.commands.base import BaseCommand
from jig.gitutils.hooking import (
    create_auto_init_templates, set_templates_directory)

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Make Jig auto-init every time you git clone',
    usage='jig sticky [-h]')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        with self.out() as out:
            templates_directory = create_auto_init_templates(expanduser('~'))

            set_templates_directory(templates_directory)

            out.append('Jig has been setup to run everytime you clone.')



"""
Git object converter
====================

`GitPython`_ does an excellent job of presenting native Git objects in Python
objects that can be manipulated and used. However, the format is not ideal for
our purposes.

What is useful to a jig plugin author is the following:

    1. What files changed
    2. What's the simple diff for modified files

This module manipulates :py:class:`git.DiffIndex` objects and provides other
utilities for discovering differences between two strings.

.. _GitPython: https://github.com/gitpython-developers/GitPython
"""
from os.path import islink
from difflib import SequenceMatcher

from git.exc import BadObject
from jig.conf import CODEC


def _make_unicode(string):
    """
    Force a conversion to unicode if necessary.

    :param str string: basestring
    :rtype: unicode
    """
    if isinstance(string, unicode):
        return string
    return string.decode(CODEC)


def describe_diff(a, b):
    """
    Takes two strings and calculates the difference between them.

    Output format is a list of::

        (line_number, diff_type, line)

    ``diff_type`` can be:
        * ``' '``
        * ``'-'``
        * ``'+'``

    Example::

        >>> list(describe_diff('a\\nb\\nc', 'a\\nc\\nd'))
        [(1, ' ', 'a'),
        (2, '-', 'b'),
        (2, ' ', 'c'),
        (3, '+', 'd')]
    """
    a = a.splitlines()
    b = b.splitlines()

    for tag, i1, i2, j1, j2 in SequenceMatcher(None, a, b).get_opcodes():
        if tag == 'equal':
            for idx, line in enumerate(b[j1:j2]):
                yield (idx + j1 + 1, ' ', _make_unicode(line))
            continue
        if tag == 'replace' or tag == 'delete':
            for idx, line in enumerate(a[i1:i2]):
                yield (idx + i1 + 1, '-', _make_unicode(line))
        if tag == 'replace' or tag == 'insert':
            for idx, line in enumerate(b[j1:j2]):
                yield (idx + j1 + 1, '+', _make_unicode(line))


class DiffType(object):

    """
    What kind of operation does this diff represent.

    """
    A = 'added'
    D = 'deleted'
    R = 'renamed'
    M = 'modified'
    U = 'unknown'

    @classmethod
    def for_diff(cls, diff):
        """
        Determines what type of diff this represents
        """
        if diff.new_file:
            return cls.A
        elif diff.deleted_file:
            return cls.D
        elif diff.renamed:
            return cls.R
        elif diff.a_blob and diff.b_blob and diff.a_blob != diff.b_blob:
            return cls.M

        return cls.U


class GitDiffIndex(object):

    """
    Converts diff index object to something useful for pre-commit hooks.

    The expected argument when creating an instance is a
    :py:class:`git.diff.DiffIndex` object.

    The following information is extracted from the list

    """
    def __init__(self, gitrepo, difflist):
        """
        Where ``gitrepo`` is the path to the root of the Git repository.
        """
        self.gitrepo = gitrepo
        self.difflist = difflist

    def files(self):
        """
        A generator for returning human-readable information about the diffs.

        Generator yields dictionaries that contain filename, name, diff, and
        type.

        ``filename`` is the absolute path to the file that was modified.

        ``name`` is the path of the file relative to the Git repository.

        ``diff`` is a list of ``(line_number, '+'|'-'|' ', line)`` describing
        the changes that occurred between the a_blob and b_blob from the
        commit.

        ``type`` is ``added``, ``deleted``, ``renamed``, ``modified`` and
        describes the overall action that occurred on this file.

        This will skip symlinks and will not provide the contens of binary
        files.
        """
        for diff in self.difflist:
            a_blob = diff.a_blob
            b_blob = diff.b_blob

            a_data = ''
            b_data = ''

            try:
                a_data = a_blob.data_stream.read()
            except (AttributeError, BadObject):
                pass

            try:
                b_data = b_blob.data_stream.read()
            except (AttributeError, BadObject):
                pass

            if '\0' in a_data or '\0' in b_data:
                # This file is binary? Probably.
                linediff = []
            else:
                linediff = describe_diff(a_data, b_data)

            blob = a_blob or b_blob

            if islink(blob.abspath):
                # Skip symlinks
                continue

            if blob.path.startswith('.jig'):
                # This is a file that is part of .jig, ignore it
                continue

            yield {
                'filename': blob.abspath,
                'name': blob.path,
                'diff': linediff,
                'type': DiffType.for_diff(diff)}



from textwrap import dedent


def _hint(hint):
    return [u''] + dedent(hint).strip().splitlines()

AFTER_INIT = _hint(
    u"""
    You should tell Git to ignore the new .jig directory. Run this:

        $ echo ".jig" >> .gitignore

    Next install some plugins. Jig has a common set you may like:

        $ curl -L https://raw.github.com/robmadole/jig-plugins/lists/common.txt > .jigplugins.txt
        $ jig install .jigplugins.txt
    """)

PRE_COMMIT_EXISTS = _hint(
    u"""
    For Jig to operate automatically when you commit we need to create
    a new pre-commit hook.

    Check the existing pre-commit file and see if you are using it.

    If you do not need the existing pre-commit script, you can delete it
    and then run jig init again in this repository.
    """)

NOT_GIT_REPO = _hint(
    u"""
    To initialize a directory for use with Git, run:

        $ git init [directory]

    After you do this you can try your Jig command again.
    """)

GIT_REPO_NOT_INITIALIZED = _hint(
    u"""
    To use Jig you must first make sure the directory has been
    initialized for use with Git.

    Make sure you run this first:

        $ git init [directory]

    Then run the init command using Jig:

        $ jig init [directory]
    """)

GIT_REV_LIST_FORMAT_ERROR = _hint(
    u"""
    The revision range is not in a valid format.

    Use "REV_A..REV_B" to specify the revisions that Jig should operate against.
    """)

GIT_REV_LIST_MISSING = _hint(
    u"""
    The revision specified is formatted correctly but one of both of the revisions
    could not be found.
    """)

GIT_TEMPLATES_MISSING = _hint(
    u"""
    Jig looks for the shared Git templates in /usr/share and /usr/local. You may
    have a non-standard installation of Git.
    """)

GIT_HOME_TEMPLATES_EXISTS = _hint(
    u"""
    You may have already configured Jig for this. Delete the directory
    and run the command again if you are sure this is what you want to
    do.
    """)

INIT_TEMPLATE_DIR_ALREADY_SET = _hint(
    u"""
    Your Git config already has a value set for init.templatedir. Jig will not
    change this because you have probably done this for a reason. If you'd like
    to use hooks from Jig you can manually copy them from ~/.jig/git/templates/hooks.
    """)

ALREADY_INITIALIZED = _hint(
    u"""
    You are initializing a Git repository for use with Jig but it
    seems this has already been done.

    If you need to re-initialize for some reason you can manually remove
    the .jig directory in the root of the repository. This is probably not
    necessary and will delete all your settings and installed plugins.
    """)

NO_PLUGINS_INSTALLED = _hint(
    u"""
    You can add plugins one at a time by running:

        $ jig plugin add URL|URL@BRANCH|PATH

    You can also install a list of plugins from a file:

        $ curl -L https://raw.github.com/robmadole/jig-plugins/lists/common.txt > .jigplugins.txt
        $ jig install .jigplugins.txt

    It's a good idea to add .jigplugins.txt to your Git repository after you are done.
    """)

USE_RUNNOW = _hint(
    u"""
    Run the plugins in the current repository with this command:

        $ jig runnow

    Jig works off of your staged files in the Git repository index.
    You place things in the index with `git add`. You will need to stage
    some files before you can run Jig.
    """)

CHANGE_PLUGIN_SETTINGS = _hint(
    u"""
    Plugin settings can be changed with the following command:

        $ jig config set BUNDLE.PLUGIN.KEY VALUE

    BUNDLE is the bundle name of an installed plugin
    PLUGIN is the name of an installed plugin.
    KEY is the name/key of the setting.
    VALUE is the desired value for the KEY.
    """)

FORK_PROJECT_GITHUB = _hint(
    u"""
    You can fork this project on GitHub.

    http://github.com/robmadole/jig
    """)

INVALID_RANGE = _hint(
    u"""
    Ranges are formatted as a number followed by two dots and another number.
    The first number specifies the start of the range. The second number
    specifies the end.

    To specify a range between 1 and 2 you would use "1..2".

    You can also specify larger ranges like "3..7"
    """)

INVALID_CONFIG_KEY = _hint(
    u"""
    Config keys must be in the following format:

    BUNDLE.PLUGIN.SETTING

    For example, if a plugin named "doublecheck" was installed from a bundle named
    "acme" the config key would be something like:

    acme.doublecheck.ignore_widgets

    For a full list of settings:

        $ jig config list
    """)



from os import rmdir
from os import access, X_OK
from os.path import join, realpath, isdir, isfile, expanduser
from tempfile import mkdtemp

import git
from mock import patch, Mock

from jig.tests.testcase import JigTestCase, result_with_hint
from jig.commands.hints import GIT_REPO_NOT_INITIALIZED
from jig.exc import (
    NotGitRepo, PreCommitExists, GitTemplatesMissing,
    JigUserDirectoryError, GitHomeTemplatesExists, InitTemplateDirAlreadySet,
    GitConfigError)
from jig.gitutils.hooking import (
    hook, _git_templates, create_auto_init_templates, set_templates_directory)


class TestHook(JigTestCase):

    """
    Can we properly hook a Git repository

    """
    def setUp(self):
        super(TestHook, self).setUp()

        # Where will our pre-commit file be?
        self.pc_filename = realpath(
            join(self.gitrepodir, '.git', 'hooks', 'pre-commit')
        )

    def test_not_git_directory(self):
        """
        The directory given is not a repo.
        """
        badrepo = mkdtemp()

        try:
            with self.assertRaises(NotGitRepo):
                hook(badrepo)
        finally:
            rmdir(badrepo)

    def test_will_stop_if_pre_commit_exists(self):
        """
        Stop if .git/hooks/pre-commit exists
        """
        # Let's create a pre-commit hook, which we would use
        with open(self.pc_filename, 'w') as fh:
            fh.write('#!/bin/sh')

        with self.assertRaises(PreCommitExists):
            hook(self.gitrepodir)

    def test_successfully_hooks(self):
        """
        Creates the pre-commit hook.
        """
        pc_filename = hook(self.gitrepodir)

        self.assertEqual(self.pc_filename, pc_filename)

    def test_hook_runs(self):
        """
        New hook will run.
        """
        pc_filename = hook(self.gitrepodir)

        retcode, output = self.runcmd(pc_filename)

        self.assertEqual(1, retcode)
        self.assertResults(
            result_with_hint(
                u'This repository has not been initialized.',
                GIT_REPO_NOT_INITIALIZED),
            output)


class TestGitTemplates(JigTestCase):

    """
    Function that finds the shared Git templates.

    """
    def test_will_find_one(self):
        """
        Can find a shared Git directory.
        """
        self.assertIsNotNone(_git_templates())

    def test_returns_none_if_not_found(self):
        """
        Returns None if it cannnot find any templates.
        """
        with patch('jig.gitutils.hooking.isdir') as isdir:
            isdir.return_value = False

            self.assertIsNone(_git_templates())


class TestCreateAutoInitTemplates(JigTestCase):

    """
    An auto-init templates directory can be created.

    """
    def setUp(self):
        super(TestCreateAutoInitTemplates, self).setUp()

        self.user_home_directory = mkdtemp()
        self.git_templates = \
            'jig.gitutils.hooking._git_templates'

    def test_raises_exception_permission_denied(self):
        """
        Raise if the .jig directory can't be created due to permissions.
        """
        with patch('jig.gitutils.hooking.makedirs') as makedirs:
            makedirs.side_effect = OSError(13, 'Permission denied')

            with self.assertRaises(JigUserDirectoryError) as ec:
                create_auto_init_templates(self.user_home_directory)

        self.assertEqual(
            u'Cannot create {0}/.jig Jig user directory'.format(
                self.user_home_directory),
            unicode(ec.exception)
        )

    def test_raises_exception_other_os_error(self):
        """
        Raise if some other kind of OSError is found.
        """
        with patch('jig.gitutils.hooking.makedirs') as makedirs:
            makedirs.side_effect = OSError(99, 'Flooglehorn is blocked')

            with self.assertRaises(JigUserDirectoryError) as ec:
                create_auto_init_templates(self.user_home_directory)

        self.assertEqual(
            u'[Errno 99] Flooglehorn is blocked',
            unicode(ec.exception)
        )

    def test_continues_if_jig_user_directory_created(self):
        """
        An existing ~/.jig directory doesn't cause failure.
        """
        with patch('jig.gitutils.hooking.makedirs') as makedirs:
            makedirs.side_effect = OSError(17, 'Directory exists')

        self.assertEqual(
            '{0}/.jig/git/templates'.format(self.user_home_directory),
            create_auto_init_templates(self.user_home_directory)
        )

    def test_raises_exception_if_no_templates(self):
        """
        If no Git templates can be found raise an exception.
        """
        with patch(self.git_templates) as gtsl:
            gtsl.return_value = None

            with self.assertRaises(GitTemplatesMissing):
                create_auto_init_templates(self.user_home_directory)

    def test_creates_the_templates_directory(self):
        """
        The templates are copied from the Git shared directory.
        """
        home_templates_directory = create_auto_init_templates(
            self.user_home_directory
        )

        self.assertTrue(isdir(home_templates_directory))

    def test_already_created(self):
        """
        Will raise an exception when it's already created the templates.
        """
        create_auto_init_templates(self.user_home_directory)

        with self.assertRaises(GitHomeTemplatesExists):
            create_auto_init_templates(self.user_home_directory)

    def test_creates_pre_commit_hook(self):
        """
        Creates the pre-commit hook for auto-initialization.
        """
        templates_directory = create_auto_init_templates(
            self.user_home_directory
        )

        self.assertTrue(
            isfile(join(templates_directory, 'hooks', 'pre-commit'))
        )

    def test_pre_commit_hook_is_executable(self):
        """
        Once created it can be executed.
        """
        templates_directory = create_auto_init_templates(
            self.user_home_directory
        )

        self.assertTrue(
            access(join(templates_directory, 'hooks', 'pre-commit'), X_OK)
        )


class TestSetTemplatesDirectory(JigTestCase):

    """
    The Git config can be changed to use the Jig templates directory.

    """
    def setUp(self):
        super(TestSetTemplatesDirectory, self).setUp()

        self.templates_directory = create_auto_init_templates(mkdtemp())

        self._clear_gitconfig()

    @classmethod
    def tearDownClass(cls):
        cls._clear_gitconfig()

    @classmethod
    def _clear_gitconfig(self):
        # Reset the global Git config
        with open(expanduser('~/.gitconfig'), 'w') as fh:
            fh.write('')

    def test_error_reading_git_config(self):
        """
        Raises an exception if there are problems reading the Git config.
        """
        with patch('jig.gitutils.hooking.git.cmd') as mock_cmd:
            mock_command = Mock()
            mock_command.config.side_effect = git.exc.GitCommandError(
                'git config', 1, 'error'
            )

            mock_cmd.Git.return_value = mock_command

            with self.assertRaises(GitConfigError) as gce:
                set_templates_directory(self.templates_directory)

        self.assertEqual(
            u'Problem when running git config: error',
            unicode(gce.exception)
        )

    def test_init_templatedir_already_set(self):
        """
        Will raise an exception if the config already has init.templatedir set.
        """
        git.cmd.Git().config(
            '--global', '--add', 'init.templatedir', '/tmp/templates'
        )

        with self.assertRaises(InitTemplateDirAlreadySet):
            set_templates_directory(self.templates_directory)

    def test_sets_templatedir(self):
        """
        Will set the templatedir.
        """
        set_templates_directory(self.templates_directory)

        config = git.cmd.Git().config('--global', '--list')

        self.assertIn(
            'init.templatedir',
            config
        )

        self.assertIn(
            self.templates_directory,
            config
        )

    def test_raises_exception_if_cannot_write(self):
        """
        Raises an exception if it can't write to the Git config.
        """
        side_effects = [
            '',
            git.exc.GitCommandError('', 1, 'error')
        ]

        def config(*args, **kwargs):
            se = side_effects.pop(0)
            if isinstance(se, Exception):
                raise se
            else:
                return se

        with patch('jig.gitutils.hooking.git.cmd') as mock_cmd:
            mock_command = Mock()
            mock_command.config.side_effect = config
            mock_cmd.Git.return_value = mock_command

            with self.assertRaises(GitConfigError):
                set_templates_directory(self.templates_directory)



from jig.commands.base import BaseCommand
from jig.runner import Runner

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Run plugins on a revision range',
    usage='jig report [-h] [-p PLUGIN] [--rev-range REVISION_RANGE] [PATH]')

_parser.add_argument(
    '--plugin', '-p',
    help='Only run this specific named plugin')
_parser.add_argument(
    '--rev-range', dest='rev_range', default='HEAD^1..HEAD',
    help='Git revision range to run the plugins against')
_parser.add_argument(
    'path', nargs='?', default='.',
    help='Path to the Git repository')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        path = argv.path
        rev_range = argv.rev_range

        # Make the runner use our view
        runner = Runner(view=self.view)

        runner.main(
            path,
            plugin=argv.plugin,
            rev_range=rev_range,
            interactive=False
        )



"""
Collection of :py:class:`Exception` subclasses used by jig.
"""


class JigException(Exception):

    """
    Base exception.

    """
    def __init__(self, *args, **kwargs):
        hint = kwargs.pop('hint', None)

        super(JigException, self).__init__(*args, **kwargs)

        if hint is not None:
            self.hint = hint


class ForcedExit(JigException):

    """
    Raised when execution of should stop.

    This exception is useful in testing. Typically the runner is configured to
    not ``sys.exit()`` on exception so this object can be used in its place.

    """
    def __eq__(self, other):
        return str(self) == str(other)


class NotGitRepo(JigException):

    """
    A directory provided does not appear to be a Git repository.

    """
    hint = 'NOT_GIT_REPO'


class PreCommitExists(JigException):

    """
    The :file:`pre-commit` file in the hooks directory exists.

    """
    hint = 'PRE_COMMIT_EXISTS'


class GitRepoNotInitialized(JigException):

    """
    The given Git repository has not been initialized for usage.

    """
    hint = 'GIT_REPO_NOT_INITIALIZED'


class GitCloneError(JigException):

    """
    Attempting to clone a Git repository failed.

    """
    pass


class GitRevListFormatError(JigException):

    """
    The given Git revision list is not in a valid format.

    """
    hint = 'GIT_REV_LIST_FORMAT_ERROR'


class GitRevListMissing(JigException):

    """
    The revision lists could not be found for a given Git repository.

    """
    hint = 'GIT_REV_LIST_MISSING'


class GitTemplatesMissing(JigException):

    """
    The shared Git templates directory is not in any expected locations.

    """
    hint = 'GIT_TEMPLATES_MISSING'

    def __init__(self):
        super(GitTemplatesMissing, self).__init__(
            'Unable to find templates.')


class GitHomeTemplatesExists(JigException):

    """
    The ~/.jig/templates directory already exists.

    """
    hint = 'GIT_HOME_TEMPLATES_EXISTS'

    def __init__(self, home_templates_directory):
        super(GitHomeTemplatesExists, self).__init__(
            '{0} already exists'.format(home_templates_directory)
        )


class InitTemplateDirAlreadySet(JigException):

    """
    If the Git config already has a value for the init.templatedir key.

    """
    hint = 'INIT_TEMPLATE_DIR_ALREADY_SET'

    def __init__(self, init_templatedir):
        super(InitTemplateDirAlreadySet, self).__init__(
            'Git configuration for init.templatedir is {0}'.format(
                init_templatedir))


class GitConfigError(JigException):

    """
    Raise when an error occurs reading or writing the Git config.

    """
    def __init__(self, git_command_error):
        """
        Use the stderr message from the command as this exception's message.

        :param git.exc.GitCommandError git_command_error:
        """
        super(GitConfigError, self).__init__(
            'Problem when running {0}: {1}'.format(
                git_command_error.command, git_command_error.stderr))


class AlreadyInitialized(JigException):

    """
    The given Git repository has been initialized already.

    """
    hint = 'ALREADY_INITIALIZED'


class JigUserDirectoryError(JigException):

    """
    An error finding or creating the ~/.jig directory.

    """
    pass


class PluginError(JigException):

    """
    An error occured while working with a plugin.

    """
    pass


class CommandError(JigException):

    """
    Raise when an error occurs while using any command line tool.

    """
    pass


class ConfigKeyInvalid(JigException):

    """
    Raise whenever the dot-separated config key is not valid.
    """
    hint = 'INVALID_CONFIG_KEY'


class RangeError(JigException):

    """
    Raise when a given numbered test range is invalid.

    """
    hint = 'INVALID_RANGE'


class ExpectationError(JigException):

    """
    Base class for all error dealing with plugin expectations.

    """
    pass


class ExpectationNoTests(ExpectationError):

    """
    Raise when no tests (numbered directories) are found for a plugin.

    """
    pass


class ExpectationFileNotFound(ExpectationError):

    """
    Raised when the expectation file should be present but is not.

    """
    pass


class ExpectationParsingError(ExpectationError):

    """
    Raised when errors are found while parsing reStructuredText expectations.

    """
    pass



import errno

from jig.commands.base import (
    BaseCommand, add_plugin, plugins_by_bundle, plugins_by_name)
from jig.commands.hints import (
    NO_PLUGINS_INSTALLED, USE_RUNNOW, FORK_PROJECT_GITHUB)
from jig.exc import CommandError, ExpectationError
from jig.tools import indent
from jig.plugins import (
    get_jigconfig, set_jigconfig, PluginManager,
    create_plugin, available_templates)
from jig.plugins.tools import update_plugins
from jig.plugins.testrunner import (
    PluginTestRunner, PluginTestReporter,
    FailureResult, parse_range)

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Manage this repository\'s Jig plugins',
    usage='jig plugin [-h] ACTION')

_subparsers = _parser.add_subparsers(
    title='actions',
    description='available commands to manage plugins')

_listparser = _subparsers.add_parser(
    'list', help='list installed plugins',
    usage='jig plugin list [-h] [-r GITREPO]')
_listparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_listparser.set_defaults(subcommand='list')

_addparser = _subparsers.add_parser(
    'add', help='add a plugin',
    usage='jig plugin add [-h] [-r GITREPO] URL|URL@BRANCH|PATH')
_addparser.add_argument(
    'plugin',
    help='URL or path to the plugin directory, if URL you can '
    'specify @BRANCHNAME to clone other than the default')
_addparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_addparser.set_defaults(subcommand='add')

_updateparser = _subparsers.add_parser(
    'update', help='update all installed plugins',
    usage='jig plugin update [-h] [-r GITREPO]')
_updateparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_updateparser.set_defaults(subcommand='update')

_removeparser = _subparsers.add_parser(
    'remove', help='remove an installed plugin',
    usage='jig plugin remove [-h] [-r GITREPO] NAME [BUNDLE]')
_removeparser.add_argument(
    'name', help='Plugin name')
_removeparser.add_argument(
    'bundle', nargs='?', default=None,
    help='Bundle name')
_removeparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_removeparser.set_defaults(subcommand='remove')

_createparser = _subparsers.add_parser(
    'create', help='create a new plugin',
    usage='jig plugin create [-h] [-l TEMPLATE] [-d DIR] NAME BUNDLE')
_createparser.add_argument(
    'name', help='Plugin name')
_createparser.add_argument(
    'bundle', help='Bundle name')
_createparser.add_argument(
    '--language', '-l', dest='template',
    default='python', help='Scripting language: {0}'.format(
        ', '.join(available_templates())))
_createparser.add_argument(
    '--dir', '-d', default='.',
    help='Create in this directory')
_createparser.set_defaults(subcommand='create')

_testparser = _subparsers.add_parser(
    'test', help='run a suite of plugin tests',
    usage='jig plugin test [-h] [-r RANGE] PLUGIN')
_testparser.add_argument(
    'plugin', nargs='?', default='.',
    help='Path to the plugin directory')
_testparser.add_argument(
    '--verbose', '-v',
    default=False, action='store_true',
    help='Print the input and output (stdin and stdout)')
_testparser.add_argument(
    '--range', '-r',
    dest='range',
    help='Run a subset of the tests, specified like [s]..[e]. Example -r '
    '3..5 to run tests that have expectations for those changes.')
_testparser.set_defaults(subcommand='test')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        subcommand = argv.subcommand

        # Handle the actions
        getattr(self, subcommand)(argv)

    def list(self, argv):
        """
        List the installed plugins.
        """
        path = argv.path

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            bundles = plugins_by_bundle(pm)

            if not bundles:
                out.append(u'No plugins installed.')
                out.extend(NO_PLUGINS_INSTALLED)
                return

            out.append(u'Installed plugins\n')

            out.append(u'{h1:<25} {h2}'.format(
                h1=u'Plugin name', h2=u'Bundle name'))

            sort_bundles = sorted(bundles.items(), key=lambda b: b[0])

            for name, plugins in sort_bundles:
                sort_plugins = sorted(plugins, key=lambda p: p.name)

                for plugin in sort_plugins:
                    out.append(u'{plugin:.<25} {name}'.format(
                        name=name, plugin=plugin.name))

            out.extend(USE_RUNNOW)

    def add(self, argv):
        """
        Add a plugin.
        """
        path = argv.path
        plugin = argv.plugin

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            added = add_plugin(pm, plugin, path)

            set_jigconfig(path, pm.config)

            for p in added:
                out.append(
                    'Added plugin {0} in bundle {1} to the '
                    'repository.'.format(p.name, p.bundle))

            out.extend(USE_RUNNOW)

    def update(self, argv):
        """
        Updates any plugins installed through a URL.

        This basically runs ``git pull`` within any directory inside the
        :file:`.jig/plugis` directory. It's a very simple method of updating
        plugins that have already been installed.
        """
        path = argv.path

        with self.out() as out:
            # Make sure that this directory has been initialized for Jig
            get_jigconfig(path)

            results = update_plugins(path)

            if not results:
                out.append('No plugins to update.')
                return

            out.append('Updating plugins')
            out.append('')

            for pm, output in results.items():
                names = set([i.name for i in pm.plugins])
                bundles = set([i.bundle for i in pm.plugins])

                out.append('Plugin {0} in bundle {1}'.format(
                    ', '.join(names), ', '.join(bundles)))
                out.extend(indent(output.splitlines()))

    def remove(self, argv):
        """
        Remove a plugin.

        This method is smart enough to work with only the plugin name if it
        happens to be unique. If there is more than one plugin with the same
        name but in a different bundle it will exit with an error.
        """
        path = argv.path
        name = argv.name
        bundle = argv.bundle

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            plugins = plugins_by_name(pm)

            # Find the bundle if it's not specified
            if name in plugins and not bundle:
                if len(plugins[name]) > 1:
                    # There are more than one plugin by this name
                    raise CommandError(
                        'More than one plugin has the name of '
                        '{0}. Use the list command to see installed '
                        'plugins.'.format(name))

                bundle = plugins[name][0].bundle

            pm.remove(bundle, name)

            set_jigconfig(path, pm.config)

            out.append('Removed plugin {0}'.format(name))

    def create(self, argv):
        """
        Create a new plugin.
        """
        name = argv.name
        bundle = argv.bundle
        template = argv.template
        save_dir = argv.dir

        with self.out() as out:
            if template not in available_templates():
                raise CommandError(
                    'Language {0} is not supported yet.'.format(template),
                    hint=FORK_PROJECT_GITHUB)

            try:
                plugin_dir = create_plugin(
                    save_dir, bundle, name,
                    template=template)

                out.append('Created plugin as {0}'.format(plugin_dir))
            except OSError as ose:
                if ose.errno == errno.EEXIST:
                    # File exists
                    raise CommandError(
                        'A plugin with this name already '
                        'exists in this directory: {0}.'.format(save_dir))
                # Something else, raise it again
                raise ose   # pragma: no cover

    def test(self, argv):
        """
        Run the tests for a plugin.
        """
        plugin = argv.plugin
        test_range = argv.range
        verbose = argv.verbose

        with self.out() as out:
            if test_range:
                test_range = parse_range(test_range)

            try:
                ptr = PluginTestRunner(plugin)

                results = ptr.run(test_range=test_range)

                reporter = PluginTestReporter(results)

                test_results = reporter.dumps(verbose=verbose).splitlines()

                failures = [i for i in results if isinstance(i, FailureResult)]

                if failures:
                    # Raise as an error so the status code will be non-zero
                    raise CommandError('\n'.join(test_results))

                # No failures, ok to send this to stdout
                out.extend(test_results)
            except ExpectationError as e:
                raise CommandError(str(e))



from mock import Mock

from jig.tests.testcase import JigTestCase
from jig.tests.mocks import MockPlugin


class TestMockPlugin(JigTestCase):

    """
    Mock plugin behaves correctly.

    """
    def test_create_unnamed_mock_plugin(self):
        """
        Create an unnamed mock plugin.
        """
        mp = MockPlugin()

        self.assertEqual('Unnamed', mp.name)

    def test_create_named_plugin(self):
        """
        Create a mock plugin with a specific name.
        """
        mp = MockPlugin(name='Specific')

        self.assertEqual('Specific', mp.name)

    def test_change_name_later(self):
        """
        Change a plugins name after it has been created.
        """
        mp = MockPlugin()

        mp.name = 'Specific'

        self.assertEqual('Specific', mp.name)

    def test_children_are_normal_mocks(self):
        """
        Calling an attribute returns a normal :py:class:`Mock`.
        """
        mp = MockPlugin()

        mock = mp.testattr

        self.assertTrue(isinstance(mock, Mock))
        self.assertFalse(isinstance(mock, MockPlugin))



import sys


def main():
    """
    Main jig command line tool.

    Provides access to all the jig user commands.
    """
    from jig.runner import Runner

    jig = Runner()
    jig.fromconsole(sys.argv)


def test():
    """
    Run the suite of tests for jig.
    """
    import nose

    from jig.tests.noseplugin import TestSetup

    nose.main(argv=['nose', '-w', 'src'] + sys.argv[1:],
            addplugins=[TestSetup()])


def coverage():
    """
    Create console and HTML coverage reports for the full test suite.
    """
    import nose
    from coverage import coverage

    from jig.tests.noseplugin import TestSetup

    omit = [
        '*noseplugin*',
        '*entrypoints*',
        '*testcase*',
        '*packages*',
        '*backports.py',
        '*jig/__init__.py']

    cov = coverage(
        branch=True, config_file=False, source=['jig'],
        omit=omit)

    cov.start()

    nose.run(argv=['nose', '-w', 'src'] + sys.argv[1:],
             addplugins=[TestSetup()])

    cov.stop()

    cov.report()
    cov.html_report(directory='../cover')



# coding=utf-8
from os.path import dirname, isdir, isfile, join
from os import makedirs
from tempfile import mkdtemp

from mock import Mock, patch

from jig.tests.testcase import (
    CommandTestCase, PluginTestCase,
    cd_gitrepo, cwd_bounce, result_with_hint)
from jig.tests.mocks import MockPlugin
from jig.tools import NumberedDirectoriesToGit
from jig.exc import ForcedExit
from jig.plugins import (
    set_jigconfig, get_jigconfig, create_plugin,
    PluginManager)
from jig.plugins.testrunner import (
    Expectation, SuccessResult,
    FailureResult, REPORTER_HORIZONTAL_DIVIDER)
from jig.gitutils.remote import clone
from jig.commands import plugin
from jig.commands.hints import (
    FORK_PROJECT_GITHUB, NO_PLUGINS_INSTALLED, USE_RUNNOW, INVALID_RANGE)


class TestPluginCommand(CommandTestCase, PluginTestCase):

    """
    Test the plugin command.

    """
    command = plugin.Command

    def setUp(self):
        super(TestPluginCommand, self).setUp()

        self.plugindir = mkdtemp()

    def _add_plugin(self, plugin_dir):
        """
        Adds a plugin to the jig initialized Git repository.
        """
        config = get_jigconfig(self.gitrepodir)
        pm = PluginManager(config)
        pm.add(plugin_dir)
        set_jigconfig(self.gitrepodir, pm.config)

    @cd_gitrepo
    def test_list_no_plugins(self):
        """
        No plugins are installed and the list is empty.
        """
        # Without the path argument, this should default to the current working
        # directory.
        self.run_command('list')

        self.assertResults(
            result_with_hint(
                u'No plugins installed.',
                NO_PLUGINS_INSTALLED),
            self.output)

    def test_list_plugins_different_bundle(self):
        """
        Lists plugins correctly if they are in different bundles.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01'))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test02', name='plugin02'))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test03', name='plugin03'))

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(result_with_hint(u'''
            Installed plugins

            Plugin name               Bundle name
            plugin01................. test01
            plugin02................. test02
            plugin03................. test03
            ''', USE_RUNNOW), self.output)

    def test_list_plugins_same_bundle(self):
        """
        Lists plugins correctly if they are in the same bundle.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test', name='plugin01'))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test', name='plugin02'))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test', name='plugin03'))

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(result_with_hint(u'''
            Installed plugins

            Plugin name               Bundle name
            plugin01................. test
            plugin02................. test
            plugin03................. test
            ''', USE_RUNNOW), self.output)

    def test_lists_alphabetically(self):
        """
        Will list bundles and plugin names alphabetically.
        """
        # Add these in reverse order of alphabetical
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='c', name='c'))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='b', name='b'))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='a', name='a'))

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(result_with_hint(u'''
            Installed plugins

            Plugin name               Bundle name
            a........................ a
            b........................ b
            c........................ c
            ''', USE_RUNNOW), self.output)

    @cd_gitrepo
    def test_add_bad_plugin(self):
        """
        Only adds a plugin if it's valid.
        """
        # This is not a valid plugin directory, it's empty
        tmp_dir = mkdtemp()

        with self.assertRaises(ForcedExit):
            self.run_command('add {0}'.format(tmp_dir))

        self.assertRegexpMatches(
            self.error,
            u'The plugin file (.+)config.cfg is missing')

    @cd_gitrepo
    def test_add_plugin(self):
        """
        Adds a valid plugin.
        """
        plugin_dir = create_plugin(
            self.plugindir, template='python',
            bundle='a', name='a')

        # We are going to test whether it defaults --gitrepo to cwd
        self.run_command('add {0}'.format(plugin_dir))

        config = get_jigconfig(self.gitrepodir)

        # The config now contains our section
        self.assertTrue(config.has_section('plugin:a:a'))

        self.assertResults(
            u'''
            Added plugin a in bundle a to the repository.

            Run the plugins in the current repository with this command:

                $ jig runnow

            Jig works off of your staged files in the Git repository index.
            You place things in the index with `git add`. You will need to stage
            some files before you can run Jig.''',
            self.output)

    def test_add_plugin_to_git_repo(self):
        """
        Add a plugin when not inside the Git repository.
        """
        plugin_dir = create_plugin(
            self.plugindir, template='python',
            bundle='a', name='a')

        self.run_command('add --gitrepo {0} {1}'.format(
            self.gitrepodir, plugin_dir))

        self.assertResults(
            u'''
            Added plugin a in bundle a to the repository.

            Run the plugins in the current repository with this command:

                $ jig runnow

            Jig works off of your staged files in the Git repository index.
            You place things in the index with `git add`. You will need to stage
            some files before you can run Jig.''',
            self.output)

    def test_add_plugin_by_url(self):
        """
        Add a plugin from a Git URL.
        """
        def clone_fake(plugin, to_dir, branch=None):
            makedirs(to_dir)
            create_plugin(
                to_dir, template='python',
                bundle='a', name='a')

        with patch('jig.commands.base.clone') as c:
            c.side_effect = clone_fake

            self.run_command(
                'add --gitrepo {0} http://repo'.format(
                    self.gitrepodir))

        # And clone was called with our URL and would have performed the
        # operation in our test directory.
        self.assertEqual('http://repo', c.call_args[0][0])
        self.assertIn(
            '{0}/.jig/plugins/'.format(self.gitrepodir),
            c.call_args[0][1])
        self.assertEqual(None, c.call_args[0][2])

    def test_add_plugin_by_url_with_branch(self):
        """
        Add a plugin from a Git URL, targeting a specific branch.
        """
        def clone_fake(plugin, to_dir, branch=None):
            makedirs(to_dir)
            create_plugin(
                to_dir, template='python',
                bundle='a', name='a')

        with patch('jig.commands.base.clone') as c:
            c.side_effect = clone_fake

            self.run_command(
                'add --gitrepo {0} http://url.com/repo@alternate'.format(
                    self.gitrepodir))

        # And the branch name was passed to clone
        self.assertEqual('alternate', c.call_args[0][2])

    def test_update_existing_plugins(self):
        """
        Can update an existing plugin.
        """
        # Make our remote repository so we have something to pull from
        origin_repo = mkdtemp()
        root_commit_dir = join(origin_repo, '01')
        makedirs(root_commit_dir)

        # Create a plugin in the repo
        create_plugin(
            root_commit_dir, template='python',
            bundle='a', name='a')
        create_plugin(
            root_commit_dir, template='python',
            bundle='b', name='b')

        # This is the directory we will clone
        ngd = NumberedDirectoriesToGit(origin_repo)
        dir_to_clone = ngd.repo.working_dir

        # This is a trick, we give it the dir_to_clone when asked to install it
        def clone_local(plugin, to_dir, branch):
            # Instead of jumping on the Internet to clone this, we will use the
            # local numbered directory repository we setup above. This will
            # allow our update to occur with a git pull and avoid network
            # traffic which is always faster for tests.
            clone(dir_to_clone, to_dir)

        # First thing is to install the the plugin
        with patch('jig.commands.base.clone') as c:
            c.side_effect = clone_local

            self.run_command(
                'add --gitrepo {0} http://repo'.format(
                    self.gitrepodir))

        self.run_command(
            'update --gitrepo {0}'.format(
                self.gitrepodir))

        self.assertResults(
            """
            Updating plugins

            Plugin a, b in bundle a, b
                Already up-to-date.""",
            self.output)

    def test_update_existing_plugins_no_plugins(self):
        """
        If an attempt is made to update plugins when none are installed.
        """
        self.run_command('update --gitrepo {0}'.format(
            self.gitrepodir))

        self.assertResults("No plugins to update.", self.output)

    @cd_gitrepo
    def test_remove_bad_plugin(self):
        """
        Only removes a plugin that has been installed.
        """
        with self.assertRaises(ForcedExit):
            self.run_command('remove a')

        self.assertEqual(
            u'This plugin does not exist.\n',
            self.error)

    @cd_gitrepo
    def test_remove_plugin(self):
        """
        Removes an installed plugin.
        """
        plugin_dir = create_plugin(
            self.plugindir, template='python',
            bundle='bundle', name='name')

        self.run_command('add -r {0} {1}'.format(self.gitrepodir, plugin_dir))

        # Remove with the --gitrepo defaulting to cwd again
        self.run_command('remove name bundle')

        config = get_jigconfig(self.gitrepodir)

        # It should be removed from our config now
        self.assertFalse(config.has_section('plugin:bundle:name'))

        self.assertEqual(
            u'Removed plugin name\n',
            self.output)

    @cd_gitrepo
    def test_remove_plugin_guesses_bundle(self):
        """
        Removes an installed plugin.
        """
        plugin_dir = create_plugin(
            self.plugindir, template='python',
            bundle='bundle', name='name')

        self.run_command('add -r {0} {1}'.format(self.gitrepodir, plugin_dir))

        # Leave the bundle name off so it can be guessed.
        self.run_command('remove name')

        self.assertEqual(
            u'Removed plugin name\n',
            self.output)

    def test_remove_plugin_same_name(self):
        """
        Exits because more than one plugin has the same name.

        If the bundle is not specified and more than one plugin has the same
        name, we can't assume which plugin they wish to remove. Error out and
        suggest they use the list command.
        """
        plugin_dir1 = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle1', name='name')
        plugin_dir2 = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle2', name='name')

        self.run_command('add -r {0} {1}'.format(self.gitrepodir, plugin_dir1))
        self.run_command('add -r {0} {1}'.format(self.gitrepodir, plugin_dir2))

        with self.assertRaises(ForcedExit):
            # Leave the bundle out, this should make our command error out
            self.run_command('remove -r {0} name'.format(self.gitrepodir))

        self.assertEqual(
            u'More than one plugin has the name of name. Use the list '
            u'command to see installed plugins.\n',
            self.error)

    def test_create_with_bad_language(self):
        """
        Cannot create a plugin if the language is unavailable
        """
        with self.assertRaises(ForcedExit):
            # We just created a plugin in this directory, so it should fail
            self.run_command('create -l php name bundle')

        self.assertResults(
            result_with_hint(
                u'Language php is not supported yet.',
                FORK_PROJECT_GITHUB),
            self.error)

    def test_create_plugin_already_exists(self):
        """
        Cannot create a plugin if the destination already exists.
        """
        save_dir = dirname(create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name'))

        with self.assertRaises(ForcedExit):
            # We just created a plugin in this directory, so it should fail
            self.run_command('create --dir {0} name bundle'.format(save_dir))

        self.assertEqual(
            u'A plugin with this name already exists in this '
            u'directory: {0}.\n'.format(save_dir),
            self.error)

    def test_create_plugin(self):
        """
        Can create a plugin.
        """
        with cwd_bounce(self.plugindir):
            self.run_command('create name bundle')

        self.assertTrue(isdir(join(self.plugindir, 'name')))
        self.assertTrue(isfile(join(self.plugindir, 'name', 'config.cfg')))

    def test_create_plugin_in_directory(self):
        """
        Creates a plugin in a given directory.
        """
        self.run_command('create --dir {0} name bundle'.format(self.plugindir))

        self.assertTrue(isdir(join(self.plugindir, 'name')))

    def test_create_plugin_defaults_python(self):
        """
        Creates a plugin with the default language of python.
        """
        self.run_command(
            'create --dir {0} --language python name bundle'.format(
                self.plugindir))

        with open(join(self.plugindir, 'name', 'pre-commit')) as fh:
            pre_commit = fh.readlines()

        self.assertEqual('#!/usr/bin/env python\n', pre_commit[0])

    def test_plugin_tests_none_found(self):
        """
        Run tests for a plugin where no tests are found.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        with self.assertRaises(ForcedExit):
            self.run_command('test {0}'.format(plugin_dir))

        self.assertIn('Could not find any tests:', self.error)
        self.assertIn('{0}/tests'.format(plugin_dir), self.error)

    def test_formats_results(self):
        """
        Will return test results.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            SuccessResult(
                actual=u'aaa', expectation=expectation,
                plugin=MockPlugin())]

        with patch('jig.commands.plugin.PluginTestRunner') as ptr:
            ptr.return_value = Mock()
            ptr.return_value.run = Mock(return_value=results)

            self.run_command('test {0}'.format(plugin_dir))

        self.assertResults(
            u'''
            01 – 02 Pass

            Pass 1, Fail 0''', self.output)

    def test_runs_specific_test(self):
        """
        Will run a specific test.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        with patch('jig.commands.plugin.PluginTestRunner') as ptr:
            ptr.return_value = Mock()
            ptr.return_value.run = Mock(return_value=[])

            self.run_command('test -r 4..5 {0}'.format(plugin_dir))

        ptr.return_value.run.assert_called_with(test_range=[(4, 5)])

    def test_handles_range_error(self):
        """
        If an improper range is given, provides a helpful error message.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        with self.assertRaises(ForcedExit):
            # Bad range "a.b"
            self.run_command('test -r a.b {0}'.format(plugin_dir))

        self.assertResults(
            result_with_hint(
                u'a.b is an invalid numbered test range',
                INVALID_RANGE),
            self.error)

    def test_plugin_test_failure(self):
        """
        Fails with exit code other than 0.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        expectation = Expectation((1, 2), None, u'bbb')
        results = [
            FailureResult(
                actual=u'aaa', expectation=expectation,
                plugin=MockPlugin())]

        with patch('jig.commands.plugin.PluginTestRunner') as ptr:
            ptr.return_value = Mock()
            ptr.return_value.run = Mock(return_value=results)

            with self.assertRaises(ForcedExit):
                self.run_command('test {0}'.format(plugin_dir))

        self.assertResults(
            u'''
            01 – 02 Fail

            Actual
            {0}

            aaa

            Diff
            {0}

            - bbb
            + aaa

            Pass 0, Fail 1'''.format(REPORTER_HORIZONTAL_DIVIDER),
            self.error)

    def test_plugin_defaults_to_cwd(self):
        """
        Running the plugins tests defaults to the current working directory.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            SuccessResult(
                actual=u'aaa', expectation=expectation,
                plugin=MockPlugin())]

        with patch('jig.commands.plugin.PluginTestRunner') as ptr:
            ptr.return_value = Mock()
            ptr.return_value.run = Mock(return_value=results)

            with cwd_bounce(plugin_dir):
                self.run_command('test')

        self.assertResults(
            u'''
            01 – 02 Pass

            Pass 1, Fail 0''', self.output)

    def test_formats_results_verbose(self):
        """
        Will return test results with stdin and stdout.
        """
        plugin_dir = create_plugin(
            mkdtemp(), template='python',
            bundle='bundle', name='name')

        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            SuccessResult(
                actual=u'aaa', expectation=expectation,
                plugin=MockPlugin(), stdin='a\n', stdout='b\n')]

        with patch('jig.commands.plugin.PluginTestRunner') as ptr:
            ptr.return_value = Mock()
            ptr.return_value.run = Mock(return_value=results)

            self.run_command('test -v {0}'.format(plugin_dir))

        self.assertResults(
            u'''
            01 – 02 Pass

            stdin (sent to the plugin)

                a

            stdout (received from the plugin)

                b

            {0}
            Pass 1, Fail 0'''.format(REPORTER_HORIZONTAL_DIVIDER),
            self.output)



from .tools import (
    initializer, set_jigconfig, get_jigconfig, create_plugin,
    available_templates, set_checked_for_updates, last_checked_for_updates)
from .manager import PluginManager, Plugin



from os.path import join
from tempfile import mkdtemp
from shutil import rmtree
from time import sleep

from mock import patch
from git import Git, Repo
from git.exc import GitCommandError

from jig.tests.testcase import JigTestCase
from jig.exc import GitCloneError
from jig.gitutils.checks import is_git_repo
from jig.gitutils.remote import clone, remote_has_updates


class TestClone(JigTestCase):

    """
    Git utils clone method can clone a repository.

    """
    def setUp(self):
        self.workingdir = mkdtemp()

    def tearDown(self):
        rmtree(self.workingdir)

    def test_clone_valid_repo(self):
        """
        Valid repo can be cloned.
        """
        with patch.object(Git, 'execute'):
            to_dir = join(self.workingdir, 'a')

            Git.execute.return_value = 'Cloning into X'

            gitobj = clone('http://github.com/user/repo', to_dir)

            Git.execute.assert_called_with([
                'git', 'clone', 'http://github.com/user/repo', to_dir
            ])

        self.assertIsInstance(gitobj, Git)

    def test_clone_invalid_repo(self):
        """
        Invalid repo raises error.
        """
        with patch.object(Git, 'execute'):
            to_dir = join(self.workingdir, 'a')

            Git.execute.side_effect = GitCommandError(
                ['command'], 128, stderr='bad command'
            )

            with self.assertRaises(GitCloneError) as gce:
                clone('http://github.com/user/repo', to_dir)

            self.assertIn(
                "'command' returned exit status 128: bad command",
                gce.exception
            )

    def test_local_directory_clone(self):
        """
        Clones a local file-based Git repository.
        """
        to_dir = join(self.workingdir, 'a')

        clone(self.gitrepodir, to_dir)

        self.assertTrue(is_git_repo(to_dir))

    def test_clone_branch(self):
        """
        Clone a specific branch of a repository.
        """
        with patch.object(Git, 'execute'):
            to_dir = join(self.workingdir, 'a')

            Git.execute.return_value = 'Cloning into X'

            clone(
                'http://github.com/user/repo',
                to_dir,
                branch='alternate'
            )

            Git.execute.assert_called_with([
                'git', 'clone', '--branch', 'alternate',
                'http://github.com/user/repo', to_dir
            ])


class TestRemoteHasUpdates(JigTestCase):

    """
    Git utils check if the active branch is older than the remote.

    """
    def setUp(self):
        super(TestRemoteHasUpdates, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self.remote_repo = repo
        self.remote_workingdir = working_dir

        self.local_workingdir = mkdtemp()

        clone(self.remote_workingdir, self.local_workingdir)

        self.local_repo = Repo(self.local_workingdir)

    def tearDown(self):
        rmtree(self.local_workingdir)

    def test_no_updates(self):
        """
        If the remote and local are the same, return False.
        """
        self.assertFalse(remote_has_updates(self.local_workingdir))

    def test_has_updates(self):
        """
        If the remote is newer than the local, returns True.
        """
        # Wait a second so the date is different than our original commit
        sleep(1.0)

        self.commit(self.remote_workingdir, 'a.txt', 'aaa')

        self.assertTrue(remote_has_updates(self.local_workingdir))

    def test_handles_git_python_exceptions(self):
        """
        If the fetch to retrieve new information results in an exception.
        """
        with patch('jig.gitutils.remote.git') as git:
            git.Repo.side_effect = AttributeError

            self.assertTrue(remote_has_updates(self.local_workingdir))

        with patch('jig.gitutils.remote.git') as git:
            git.Repo.side_effect = GitCommandError(None, None)

            self.assertTrue(remote_has_updates(self.local_workingdir))

        with patch('jig.gitutils.remote.git') as git:
            git.Repo.side_effect = AssertionError

            self.assertTrue(remote_has_updates(self.local_workingdir))

    def test_has_updates_in_local(self):
        """
        If the updates are in the local branch, return False.
        """
        sleep(1.0)

        self.commit(self.local_workingdir, 'a.txt', 'aaa')

        self.assertFalse(remote_has_updates(self.local_workingdir))

    def test_remote_different_branch_has_updates(self):
        """
        If the remote has a non-"master" branch as the default.
        """
        # Create a new branch on the remote
        alternate = self.remote_repo.create_head('alternate')
        self.remote_repo.head.reference = alternate
        self.remote_repo.head.reset(index=True, working_tree=True)

        # Clone the remote branch locally
        self.local_workingdir = mkdtemp()
        clone(self.remote_workingdir, self.local_workingdir,
              branch='alternate')
        self.local_repo = Repo(self.local_workingdir)

        # If we check now, no new updates have been made
        self.assertFalse(remote_has_updates(self.local_workingdir))

        # Let the clock rollover
        sleep(1.0)

        # Commit to the 'alternate' branch on the remote
        self.commit(self.remote_workingdir, 'a.txt', 'aaa')

        self.assertTrue(remote_has_updates(self.local_workingdir))



# coding=utf-8
from jig.tests.testcase import JigTestCase, ViewTestCase
from jig.tests.mocks import MockPlugin
from jig.output import (
    ConsoleView, Message, Error, ResultsCollater, OK_SIGN, ATTENTION, EXPLODE)

try:
    from collections import OrderedDict
except ImportError:   # pragma: no cover
    from ordereddict import OrderedDict


class TestConsoleView(ViewTestCase):

    """
    With plugin results we can format them to the console.

    """
    def setUp(self):
        self.view = ConsoleView()

        self.view.collect_output = True
        self.view.exit_on_exception = False

    def test_error(self):
        """
        The plugin exits with something other than 0.
        """
        plugin = MockPlugin()
        plugin.name = 'Plugin 1'

        counts = self.view.print_results({
            plugin: (1, '', 'An error occurred')})

        self.assertEqual((0, 0, 0), counts)
        self.assertResults(u'''
            ▾  Plugin 1

            ✕  An error occurred

            {0}  Jig ran 1 plugin
                Info 0 Warn 0 Stop 0
                (1 plugin reported errors)
            '''.format(ATTENTION), self.output)

    def test_commit_specific_message(self):
        """
        Messages generalized for the entire commit.
        """
        plugin = MockPlugin()
        plugin.name = 'Plugin 1'

        counts = self.view.print_results({
            plugin: (0, 'commit', '')})

        self.assertEqual((1, 0, 0), counts)
        self.assertResults(u"""
            ▾  Plugin 1

            ✓  commit

            {0}  Jig ran 1 plugin
                Info 1 Warn 0 Stop 0
            """.format(ATTENTION), self.output)

    def test_file_specific_message(self):
        """
        Messages specific to the file being committed.
        """
        plugin = MockPlugin()
        plugin.name = 'Plugin 1'

        counts = self.view.print_results({
            plugin: (0, {u'a.txt': [[None, u'w', 'file']]}, '')})

        self.assertEqual((0, 1, 0), counts)
        self.assertResults(u"""
            ▾  Plugin 1

            ⚠  a.txt
                file

            {0}  Jig ran 1 plugin
                Info 0 Warn 1 Stop 0
            """.format(ATTENTION), self.output)

    def test_line_specific_message(self):
        """
        Messages specific to a single line.
        """
        plugin = MockPlugin()
        plugin.name = 'Plugin 1'

        counts = self.view.print_results({
            plugin: (0, {u'a.txt': [[1, 's', 'stop']]}, '')})

        self.assertEqual((0, 0, 1), counts)
        self.assertResults(u"""
            ▾  Plugin 1

            ✕  line 1: a.txt
                stop

            {0}  Jig ran 1 plugin
                Info 0 Warn 0 Stop 1
            """.format(EXPLODE), self.output)

    def test_two_plugins(self):
        """
        Formats messages (more than one) correctly.
        """
        plugin1 = MockPlugin()
        plugin1.name = 'Plugin 1'

        plugin2 = MockPlugin()
        plugin2.name = 'Plugin 2'

        results = OrderedDict()

        results[plugin1] = (0, ['a', 'b'], '')
        results[plugin2] = (0, ['a', 'b'], '')

        counts = self.view.print_results(results)

        self.assertEqual((4, 0, 0), counts)
        self.assertResults(u"""
            ▾  Plugin 1

            ✓  a

            ✓  b

            ▾  Plugin 2

            ✓  a

            ✓  b

            {0}  Jig ran 2 plugins
                Info 4 Warn 0 Stop 0
            """.format(ATTENTION), self.output)


class TestMessage(JigTestCase):

    """
    Test Message which is responsible for representing output from plugins.

    """
    def test_representation(self):
        """
        Object's representation is correct.
        """
        message = Message(None, type='w', file='a.txt', body='body', line=1)

        self.assertEqual(
            "<Message type=\"warn\", body='body', file='a.txt', line=1>",
            repr(message))

        message = Message(None, type='stop', file='a.txt', body=True, line=1)

        self.assertEqual(
            "<Message type=\"stop\", body=True, file='a.txt', line=1>",
            repr(message))

    def test_equality(self):
        """
        Messages with the same content are considered equal.
        """
        message1 = Message(MockPlugin(),
            type='w', file='a.txt', body='body', line=1)
        message2 = Message(MockPlugin(),
            type='w', file='a.txt', body='body', line=1)
        message3 = Message(MockPlugin(),
            type='w', file='b.txt', body='bbbb', line=9)

        self.assertTrue(message1 == message2)
        self.assertFalse(message2 == message3)
        self.assertFalse(message3 == {})


class TestResultsCollater(JigTestCase):

    """
    Collate results into digestible summaries.

    """
    def test_plugin_error(self):
        """
        Results with non-zero exit codes create error messages.
        """
        results = {
            MockPlugin(): (1, '', 'Plugin failed')}

        rc = ResultsCollater(results)

        cm, fm, lm = rc.messages

        self.assertEqual(
            [Error(None, type='stop', body='Plugin failed')],
            rc.errors)

    def test_no_results(self):
        """
        We have results but no real content.
        """
        results = {
            MockPlugin(): (0, None, ''),
            MockPlugin(): (0, '', ''),
            MockPlugin(): (0, [''], ''),
            MockPlugin(): (0, [['w', '']], ''),
            MockPlugin(): (0, {u'a.txt': u''}, ''),
            MockPlugin(): (0, {u'a.txt': [[]]}, ''),
            MockPlugin(): (0, {u'a.txt': [[u'']]}, ''),
            MockPlugin(): (0, {u'a.txt': [['', u'']]}, ''),
            MockPlugin(): (0, {u'a.txt': [[None, '', u'']]}, ''),
            MockPlugin(): (0, {u'a.txt': [[1, '', u'']]}, '')}

        rc = ResultsCollater(results)

        cm, fm, lm = rc.messages

        # Should all be empty since nothing of interest was found
        self.assertEqual([], cm)
        self.assertEqual([], fm)
        self.assertEqual([], lm)

        # And we should have no errors
        self.assertEqual([], rc.errors)

    def test_commit_specific_message(self):
        """
        Results that are commit specific are collated correctly.
        """
        results = OrderedDict()
        results[MockPlugin()] = (0, 'default', '')
        results[MockPlugin()] = (0, [[u'warn', u'warning']], '')

        rc = ResultsCollater(results)

        messages, fm, lm = rc.messages

        # Defaults to info type
        self.assertEqual(Message(None, type='info', body='default'),
            messages[0])
        self.assertEqual(Message(None, type='warn', body='warning'),
            messages[1])

        # And our type counts should be 1 info 1 warning
        self.assertEqual({u'info': 1, u'warn': 1, u'stop': 0}, rc.counts)

        # The rest of these should be empty
        self.assertEqual([], fm)
        self.assertEqual([], lm)

        # And we should have no errors
        self.assertEqual([], rc.errors)

    def test_file_specific_message(self):
        """
        Results that are file-specific are collated correctly.
        """
        # Line number of None will be recognized as file-specific.
        stdout1 = {u'a.txt': [
            [None, u'warn', 'Problem with this file']]}
        # Will a length of 2 be recognized as file-specific?
        stdout2 = {u'a.txt': [
            [u'warn', 'Problem with this file']]}
        # Can we handle more than one file and different argument signatures
        # for the type?
        stdout3 = OrderedDict()
        stdout3[u'a.txt'] = [['Info A']]
        stdout3[u'b.txt'] = [[u'warn', 'Warn B']]
        stdout3[u'c.txt'] = [[u's', 'Stop C']]

        results = OrderedDict()
        results[MockPlugin()] = (0, stdout1, '')
        results[MockPlugin()] = (0, stdout2, '')
        results[MockPlugin()] = (0, stdout3, '')

        rc = ResultsCollater(results)

        self.assertEqual({u'info': 1, u'warn': 3, u'stop': 1}, rc.counts)

        cm, messages, lm = rc.messages

        for msg in messages:
            # They all lack a line number, making them file-specific.
            self.assertIsNone(msg.line)
            # But they have a file
            self.assertIsNotNone(msg.file)

        # First set's None get's recognized as file-specific
        self.assertEqual(
            Message(None, type='warn', body='Problem with this file',
                file='a.txt'),
            messages[0])

        # Second set get's recognized as a warning
        self.assertEqual(
            Message(None, type='warn', body='Problem with this file',
                file='a.txt'),
            messages[1])

        self.assertEqual(
            Message(None, type='info', body='Info A',
                file='a.txt'),
            messages[2])
        self.assertEqual(
            Message(None, type='warn', body='Warn B',
                file='b.txt'),
            messages[3])
        self.assertEqual(
            Message(None, type='stop', body='Stop C',
                file='c.txt'),
            messages[4])

        # The other messages should be empty
        self.assertEqual([], cm)
        self.assertEqual([], lm)

        # And we should have no errors
        self.assertEqual([], rc.errors)

    def test_line_specific_message(self):
        """
        Results that are line-specific are collated correctly.
        """
        stdout = OrderedDict()
        stdout[u'a.txt'] = [[1, None, 'Info A']]
        stdout[u'b.txt'] = [[2, u'warn', 'Warn B']]
        stdout[u'c.txt'] = [[3, u'stop', 'Stop C']]

        results = OrderedDict()
        results[MockPlugin()] = (0, stdout, '')

        rc = ResultsCollater(results)

        self.assertEqual({u'info': 1, u'warn': 1, u'stop': 1}, rc.counts)

        cm, fm, messages = rc.messages

        self.assertEqual(
            Message(None, type='info', body='Info A', file=u'a.txt', line=1),
            messages[0])
        self.assertEqual(
            Message(None, type='warn', body='Warn B', file=u'b.txt', line=2),
            messages[1])
        self.assertEqual(
            Message(None, type='stop', body='Stop C', file=u'c.txt', line=3),
            messages[2])

        # The other messages should be empty
        self.assertEqual([], cm)
        self.assertEqual([], fm)

        # And we should have no errors
        self.assertEqual([], rc.errors)

    def test_one_of_each(self):
        """
        One of each type of message is captured.
        """
        results = {
            MockPlugin(): (0, ['C'], ''),
            MockPlugin(): (0, {u'a.txt': u'F'}, ''),
            MockPlugin(): (0, {u'a.txt': [[1, None, u'L']]}, '')}

        rc = ResultsCollater(results)

        self.assertEqual({u'info': 3, u'warn': 0, u'stop': 0}, rc.counts)

        cm, fm, lm = rc.messages

        self.assertEqual(1, len(cm))
        self.assertEqual(1, len(fm))
        self.assertEqual(1, len(lm))

        self.assertEqual(
            Message(None, type="info", body="C"),
            cm[0])
        self.assertEqual(
            Message(None, type="info", body="F", file=u'a.txt'),
            fm[0])
        self.assertEqual(
            Message(None, type="info", body="L", file=u'a.txt', line=1),
            lm[0])

    def test_commit_specific_errors(self):
        """
        Exercise the errors related to commit specific messages.
        """
        anon_obj = object()

        results = {
            MockPlugin(): (0, anon_obj, '')}
        self.assertEqual(
            [Error(None, type='s', body=anon_obj)],
            ResultsCollater(results).errors)

        results = {
            MockPlugin(): (0, [[1, 2, 3, 4, 5]], '')}
        self.assertEqual(
            [Error(None, type='s', body=[1, 2, 3, 4, 5])],
            ResultsCollater(results).errors)

    def test_file_specific_errors(self):
        """
        Exercise the errors related to file specific messages.
        """
        anon_obj = object()

        results = {
            MockPlugin(): (0, {'a.txt': anon_obj}, '')}
        self.assertEqual(
            [Error(None, type='s', file='a.txt', body=anon_obj)],
            ResultsCollater(results).errors)

        results = {
            MockPlugin(): (0, {'a.txt': [anon_obj]}, '')}
        self.assertEqual(
            [Error(None, type='s', file='a.txt', body=anon_obj)],
            ResultsCollater(results).errors)

        results = {
            MockPlugin(): (0, {'a.txt': [1,  None]}, '')}
        self.assertEqual([
            Error(None, type='s', file='a.txt', body=1),
            Error(None, type='s', file='a.txt', body=None)],
            ResultsCollater(results).errors)

        results = {
            MockPlugin(): (0, {'a.txt': [[1, 2, 3, 4, 5]]}, '')}
        self.assertEqual(
            [Error(None, type='s', body={'a.txt': [[1, 2, 3, 4, 5]]})],
            ResultsCollater(results).errors)

    def test_empty_results(self):
        """
        Empty results do not generate empty messages.
        """
        rc1 = ResultsCollater({})

        rc2 = ResultsCollater({
            MockPlugin(): (0, [''], ''),
            MockPlugin(): (0, {u'a.txt': u'F'}, '')})

        rc3 = ResultsCollater({
            MockPlugin(): (0, ['C'], ''),
            MockPlugin(): (0, {u'a.txt': u'F'}, ''),
            MockPlugin(): (0, {u'a.txt': [[1, None, u'L']]}, '')})

        # First collator had no results
        self.assertEqual(0, len(rc1.plugins))
        self.assertEqual(0, len(rc1.reporters))

        # Second had a result with no meaningful content
        self.assertEqual(2, len(rc2.plugins))
        self.assertEqual(1, len(rc2.reporters))

        # Third had something to report for all plugins
        self.assertEqual(3, len(rc3.plugins))
        self.assertEqual(3, len(rc3.reporters))



# coding=utf-8
from os.path import join
from textwrap import dedent

from jig.exc import ForcedExit
from jig.tests.testcase import (
    CommandTestCase, PluginTestCase, cd_gitrepo, result_with_hint)
from jig.commands.hints import USE_RUNNOW
from jig.commands import install


class TestInstallCommand(CommandTestCase, PluginTestCase):

    """
    Test the install command.

    """
    command = install.Command

    def setUp(self):
        super(TestInstallCommand, self).setUp()

        self.plugin01_dir = join(self.fixturesdir, 'plugin01')
        self.plugin02_dir = join(self.fixturesdir, 'plugin02')

    @cd_gitrepo
    def test_plugins_file_does_not_exist(self):
        """
        The specified file does not exist.
        """
        with self.assertRaises(ForcedExit):
            self.run_command('badfilename.txt')

        self.assertResults(
            'No such file or directory',
            self.error)

    @cd_gitrepo
    def test_install_local_plugin(self):
        """
        Can install a single local file system plugin.
        """
        # Create a file to store the location of plugins to install
        self.commit(
            self.gitrepodir, 'jigplugins.txt',
            self.plugin01_dir)

        self.run_command('jigplugins.txt')

        self.assertResults(
            result_with_hint(dedent(
                u'''
                From {0}:
                 - Added plugin plugin01 in bundle test01
                '''.format(self.plugin01_dir)),
                USE_RUNNOW),
            self.output)

    @cd_gitrepo
    def test_skips_duplicates(self):
        """
        If a duplicate is being installed, skips it.
        """
        self.commit(
            self.gitrepodir, 'jigplugins.txt',
            '{0}\n{0}\n'.format(self.plugin01_dir))

        self.run_command('jigplugins.txt')

        self.assertResults(
            result_with_hint(dedent(
                u'''
                From {0}:
                 - Added plugin plugin01 in bundle test01
                From {0}:
                 - The plugin is already installed.
                '''.format(self.plugin01_dir)),
                USE_RUNNOW),
            self.output)

    @cd_gitrepo
    def test_plugins_has_one_error(self):
        """
        The specified exists but the first location is a bad plugin.

        Confirms that the install continues even if an error is found.
        """
        self.commit(
            self.gitrepodir, 'jigplugins.txt',
            '{0}\n{1}\n'.format(
                self.plugin02_dir, self.plugin01_dir))

        self.run_command('jigplugins.txt')

        self.assertResults(result_with_hint(dedent(
            u'''
            From {0}:
             - File contains parsing errors: {0}/config.cfg
            \t[line  2]: 'This is a bad config file that will fail to parse\\n'
            From {1}:
             - Added plugin plugin01 in bundle test01
            '''.format(self.plugin02_dir, self.plugin01_dir)), USE_RUNNOW),
            self.output)



from textwrap import dedent

RUN_JIG_SCRIPT = \
    dedent("""\
    #!{python_executable}
    from sys import path
    from os.path import dirname, join

    # Make sure that we can find the directory that jig is installed
    path.append('{jig_dir}')
    path.append('{git_python_dir}')
    path.append('{gitdb_dir}')
    path.append('{async_dir}')
    path.append('{smmap_dir}')

    from jig.runner import Runner

    # Start up the runner, passing in the repo directory
    jig = Runner()
    jig.fromhook(join(dirname(__file__), '..', '..'))
    """).strip()

AUTO_JIG_INIT_SCRIPT = \
    dedent("""\
    #!{python_executable}
    from sys import path, exit, stderr, stdout
    from os.path import dirname, realpath
    from os import unlink, environ
    from subprocess import check_call, CalledProcessError

    # Figure out where the Git directory is
    git_directory = realpath(environ.get('GIT_DIR', '.git'))
    repository_directory = dirname(git_directory)

    # Remove this pre-commit script
    unlink(__file__)

    stdout.write(u'Jig auto-install is beginning...\\n\\n')
    stdout.flush()

    try:
        check_call(u'jig init {{0}}'.format(repository_directory), shell=True)
    except (OSError, CalledProcessError) as e:
        stderr.write(
            u'Initialization of Jig failed so we are disabling\\n'
            u'the auto-install process. You need to manually install\\n'
            u'jig to use it.\\n'
        )
    else:
        stdout.write(
            u'\\nAfter you\\'ve installed some plugins run git commit again.\\n'
        )

    # Exit with non-zero so the commit can be checked with the real Jig
    # the next time a commit is attempted.
    exit(1)
    """).strip()



from os.path import dirname, join
from datetime import timedelta

## General settings

# What codec to use when dealing with unicode conversion
CODEC = 'utf_8'


## jig settings

# Name of the directory to create to hold jig data
# This will be created in the root of the Git repository
JIG_DIR_NAME = '.jig'

# Name of the file that will be created to manage a list of plugins and
# settings for them
JIG_PLUGIN_CONFIG_FILENAME = 'plugins.cfg'
JIG_PLUGIN_DIR = 'plugins'


## Plugin specific settings

# Each plugin has a configuration file, the name is
PLUGIN_CONFIG_FILENAME = 'config.cfg'

# Name of the script to run inside the plugin directory that represents
# the pre-commit script
PLUGIN_PRE_COMMIT_SCRIPT = 'pre-commit'

# Where can plugin pre-commit examples be found
PLUGIN_PRE_COMMIT_TEMPLATE_DIR = \
    join(dirname(__file__), 'data', 'pre-commits')

# How often to check for plugin updates
PLUGIN_CHECK_FOR_UPDATES = timedelta(days=5)

# The directory inside of the plugins directory that contains tests
PLUGIN_TESTS_DIRECTORY = 'tests'

# Name of the file that serves as both documentation and tests for a plugin
PLUGIN_EXPECTATIONS_FILENAME = 'expect.rst'



from os.path import isdir, join

from jig.conf import JIG_DIR_NAME


def is_git_repo(gitdir):
    """
    Returns boolean whether the directory appears to be a Git directory.
    """
    return isdir(join(gitdir, '.git'))


def repo_jiginitialized(gitdir):
    """
    Returns boolean ``True`` if ``jig init GITDIR`` has been ran.
    """
    return isdir(join(gitdir, JIG_DIR_NAME))



from jig.exc import RangeError
from jig.tests.testcase import JigTestCase
from jig.plugins.testrunner import parse_range


class TestParseRange(JigTestCase):

    """
    Parses a string range into a list.

    Valid ranges look like ``2..4`` or ``6..7``.

    """
    def test_missing_dots(self):
        """
        Raises an error if the two dots are missing.
        """
        with self.assertRaises(RangeError):
            parse_range('1.3')

    def test_invalid_numbers(self):
        """
        Checks for numbers.
        """
        with self.assertRaises(RangeError):
            parse_range('a..3')

    def test_second_number_is_greater(self):
        """
        The second number is greater than the first.
        """
        with self.assertRaises(RangeError):
            parse_range('2..1')

        with self.assertRaises(RangeError):
            parse_range('2..2')

    def test_simple_range(self):
        """
        Should result in just one test set.
        """
        self.assertEqual(
            parse_range('1..2'),
            [(1, 2)])

    def test_complex_range(self):
        """
        Should result in more than one.
        """
        self.assertEqual(
            parse_range('4..8'),
            [(4, 5), (5, 6), (6, 7), (7, 8)])

    def test_large_numbers(self):
        """
        Large numbers with zero padding on the left.
        """
        self.assertEqual(
            len(parse_range('040..050')),
            10)



from stat import S_IXUSR
from os import rmdir, stat, makedirs
from os.path import isfile, join
from tempfile import mkdtemp
from calendar import timegm
from ConfigParser import ConfigParser
from datetime import datetime, timedelta

from git import Git
from mock import patch

from jig.tests.testcase import JigTestCase, PluginTestCase, cd_gitrepo
from jig.exc import (
    NotGitRepo, AlreadyInitialized,
    GitRepoNotInitialized)
from jig.plugins import (
    initializer, get_jigconfig, set_jigconfig,
    PluginManager, create_plugin, available_templates)
from jig.plugins.tools import (
    update_plugins, last_checked_for_updates, set_checked_for_updates,
    plugins_have_updates, read_plugin_list)


class TestPluginConfig(JigTestCase):

    """
    Test plugin config file handling.

    """
    def test_not_git_repo(self):
        """
        Refuses to initialize a non-Git repo.
        """
        badrepo = mkdtemp()

        with self.assertRaises(NotGitRepo):
            initializer(badrepo)

        rmdir(badrepo)

    def test_can_initialize(self):
        """
        Can initialize the Git repo.
        """
        # Should return a ConfigParser instance
        plugins = initializer(self.gitrepodir)

        self.assertTrue(isinstance(plugins, ConfigParser))

    def test_already_initialized(self):
        """
        Catches an attempt to initialize a directory that already has been.
        """
        # Initialize it once
        initializer(self.gitrepodir)

        with self.assertRaises(AlreadyInitialized):
            # Initialize the same directory again
            initializer(self.gitrepodir)

    def test_get_config_not_initialized(self):
        """
        Attempting to get a config for a directory not initialized.
        """
        with self.assertRaises(GitRepoNotInitialized):
            get_jigconfig(self.gitrepodir)

    def test_get_config(self):
        """
        Get a config.
        """
        initializer(self.gitrepodir)

        plugins = get_jigconfig(self.gitrepodir)

        self.assertTrue(isinstance(plugins, ConfigParser))
        self.assertTrue(isfile(join(self.gitrepodir, '.jig', 'plugins.cfg')))

    def test_save_config_not_initialized(self):
        """
        Raises an error if saving a config where there is not Git repo.
        """
        with self.assertRaises(GitRepoNotInitialized):
            # It hasn't been initialized at this point, this should fail
            set_jigconfig(self.gitrepodir, config=None)

    def test_save_config(self):
        """
        Can save a config.
        """
        config = initializer(self.gitrepodir)

        config.add_section('test')
        config.set('test', 'foo', 'bar')

        set_jigconfig(self.gitrepodir, config=config)


class TestCreatePlugin(PluginTestCase):

    """
    Test the plugin creation utility.

    """
    def setUp(self):
        super(TestCreatePlugin, self).setUp()

        self.plugindir = mkdtemp()

    def test_list_available_templates(self):
        """
        List available templates for creating plugins.
        """
        self.assertEqual(['python'], available_templates())

    def test_missing_directory(self):
        """
        Will not try to create a plugin in a missing directory.
        """
        # Remove this so the create_plugin function has nowhere to go.
        rmdir(self.plugindir)

        with self.assertRaises(ValueError):
            create_plugin(
                self.plugindir, template='python',
                bundle='test', name='plugin')

    def test_creates_plugin(self):
        """
        Can create a plugin.
        """
        plugin_dir = create_plugin(
            self.plugindir, template='python',
            bundle='test', name='plugin')

        pre_commit_file = join(plugin_dir, 'pre-commit')

        # We have a pre-commit file
        self.assertTrue(isfile(pre_commit_file))

        # And it's executable
        sinfo = stat(pre_commit_file)
        self.assertTrue(S_IXUSR & sinfo.st_mode)

        config = ConfigParser()
        config.read(join(plugin_dir, 'config.cfg'))

        self.assertEqual(set(('settings', 'help', 'plugin')), set(config.sections()))
        self.assertEqual('test', config.get('plugin', 'bundle'))
        self.assertEqual('plugin', config.get('plugin', 'name'))
        self.assertEqual([], config.items('settings'))

    def test_new_plugin_compat_plugin_manager(self):
        """
        New plugins are compatible with the :py:class:`PluginManager`
        """
        plugin_dir = create_plugin(
            self.plugindir, template='python',
            bundle='test', name='plugin')

        pm = PluginManager(self.jigconfig)

        pm.add(plugin_dir)

        self.assertEqual(1, len(pm.plugins))
        self.assertEqual('plugin', pm.plugins[0].name)


class TestUpdatePlugins(PluginTestCase):

    """
    Plugins can be updated if installed via URL.

    """
    def test_update_empty_result(self):
        """
        If no plugins are installed, results are empty
        """
        # None are installed so we get an empty list
        self.assertEqual({}, update_plugins(self.gitrepodir))

    def test_update_results(self):
        """
        If we have two plugins that are updated.
        """
        plugins_dir = join(self.gitrepodir, '.jig', 'plugins')
        fake_cloned_plugin = join(plugins_dir, 'abcdef1234567890')

        makedirs(fake_cloned_plugin)

        create_plugin(fake_cloned_plugin, bundle='a', name='a')
        create_plugin(fake_cloned_plugin, bundle='b', name='b')

        with patch.object(Git, 'execute'):
            # Fake the git pull command
            mock_execute = Git.execute

            Git.execute.return_value = (0, 'Already up to date.', '')

            results = update_plugins(self.gitrepodir)

        pm, value = results.items()[0]

        # We have our two plugins from the manager
        self.assertEquals(2, len(pm.plugins))

        # And it called ``git pull`` on the repository
        mock_execute.assert_called_once_with(
            ['git', 'pull'], with_extended_output=True)


class TestPluginsHaveUpdates(PluginTestCase):

    """
    For multiple installed plugins, each can be checked for remote updates.

    """
    def setUp(self):
        plugins_dir = join(self.gitrepodir, '.jig', 'plugins')

        # Create 3 installed plugin directories. This simulates installing
        # plugins from different urls.
        for letter in 'abc':
            makedirs(join(plugins_dir, letter))

    def test_no_updates(self):
        """
        If no remote repositories have updates.
        """
        with patch('jig.plugins.tools.remote_has_updates') as rhu:
            # For each call to ``remote_has_updates``, answer False
            rhu.side_effect = [False, False, False]

            has_updates = plugins_have_updates(self.gitrepodir)

        self.assertEqual(3, rhu.call_count)
        self.assertFalse(has_updates)

    def test_has_updates(self):
        """
        If one remote repository has updates.
        """
        with patch('jig.plugins.tools.remote_has_updates') as rhu:
            # Have the last call to ``remote_has_updates`` answer True
            rhu.side_effect = [False, False, True]

            has_updates = plugins_have_updates(self.gitrepodir)

        self.assertEqual(3, rhu.call_count)
        # This time the answer is True because one plugin said it had updates
        self.assertTrue(has_updates)

    def test_all_have_updates(self):
        """
        If all repositories have updates.
        """
        with patch('jig.plugins.tools.remote_has_updates') as rhu:
            # This time they all report that they have updates
            rhu.side_effect = [True, True, True]

            has_updates = plugins_have_updates(self.gitrepodir)

        # We only need to get one True, no need to check the rest
        self.assertEqual(1, rhu.call_count)
        self.assertTrue(has_updates)


class TestCheckedForUpdates(PluginTestCase):

    """
    Determining and setting the date plugins were last checked for updates.

    """
    def test_set_last_checked(self):
        """
        Can set the date last checked.
        """
        now = timegm(datetime.utcnow().replace(microsecond=0).timetuple())
        config = set_checked_for_updates(self.gitrepodir)

        actual = int(config.get('jig', 'last_checked_for_updates'))

        self.assertEqual(actual, now)

    def test_no_last_checked(self):
        """
        If the repo has never been checked for an update.
        """
        config = get_jigconfig(self.gitrepodir)
        config.remove_section('jig')
        set_jigconfig(self.gitrepodir, config)

        last_check = last_checked_for_updates(self.gitrepodir)

        self.assertEqual(0, last_check)

    def test_bad_last_checked(self):
        """
        If the repo has a bad last checked value.
        """
        config = get_jigconfig(self.gitrepodir)
        config.set('jig', 'last_checked_for_updates', 'bad')
        set_jigconfig(self.gitrepodir, config)

        last_check = last_checked_for_updates(self.gitrepodir)

        self.assertEqual(0, last_check)

    def test_last_checked(self):
        """
        Can determine the last time checked.
        """
        now = datetime.utcnow().replace(microsecond=0)

        set_jigconfig(self.gitrepodir,
                      config=set_checked_for_updates(self.gitrepodir))

        date = last_checked_for_updates(self.gitrepodir)

        self.assertEqual(now, date)

    def test_set_last_checked_older_date(self):
        """
        Can set the date to an older value than now.
        """
        now = datetime.utcnow()
        older = now - timedelta(days=5)
        config = set_checked_for_updates(self.gitrepodir, date=older)

        now = timegm(now.timetuple())
        actual = int(config.get('jig', 'last_checked_for_updates'))
        expected = timegm(older.replace(microsecond=0).timetuple())

        self.assertEqual(actual, expected)
        self.assertGreater(now, actual)

    def test_already_checked_before(self):
        """
        If this is not the first time a check has been set.
        """
        set_jigconfig(self.gitrepodir,
                      config=set_checked_for_updates(self.gitrepodir))

        date1 = last_checked_for_updates(self.gitrepodir)

        set_jigconfig(self.gitrepodir,
                      config=set_checked_for_updates(self.gitrepodir))

        date2 = last_checked_for_updates(self.gitrepodir)

        self.assertEqual(date1, date2)


class TestReadPluginList(PluginTestCase):

    """
    Can read a file that will contain a list of plugin locations.

    """
    @cd_gitrepo
    def test_file_does_not_exist(self):
        """
        Specified file does not exist.
        """
        with self.assertRaises(IOError) as ec:
            read_plugin_list('not_a_file.txt')

        self.assertEqual(
            'No such file or directory',
            ec.exception[1])

    @cd_gitrepo
    def test_is_directory(self):
        """
        If the path is actually a directory.
        """
        self.commit(
            self.gitrepodir, 'a/b.txt',
            'foo\nbar\nbaz\n')

        with self.assertRaises(IOError) as ec:
            read_plugin_list('a')

        self.assertEqual(
            'Is a directory',
            ec.exception[1])

    @cd_gitrepo
    def test_reads_file(self):
        """
        Can read the file and return its contents.
        """
        self.commit(
            self.gitrepodir, 'a/b.txt',
            'foo\nbar\nbaz\n')

        plugin_list = read_plugin_list('a/b.txt')

        self.assertEqual(
            [u'foo', u'bar', u'baz'],
            plugin_list)



# coding=utf-8
from jig.exc import ForcedExit
from jig.tests.testcase import (
    CommandTestCase, PluginTestCase, result_with_hint)
from jig.plugins import set_jigconfig
from jig.output import ATTENTION
from jig.commands import report


class TestReportCommand(CommandTestCase, PluginTestCase):

    """
    Test the report command.

    """
    command = report.Command

    def setUp(self):
        super(TestReportCommand, self).setUp()

        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create a few commits
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.commit(self.gitrepodir, 'b.txt', 'b')
        self.commit(self.gitrepodir, 'c.txt', 'c')

    def test_cannot_resolve_revision_range(self):
        """
        Given a range that is invalid it notifies the user.
        """
        with self.assertRaises(ForcedExit) as ec:
            self.run_command('--rev-range FOO..BAR {0}'.format(self.gitrepodir))

        self.assertSystemExitCode(ec.exception, 1)

        self.assertTrue(self.error)

    def test_reports_one_commit(self):
        """
        With a range indicating one commit it reports on that one.
        """
        with self.assertRaises(SystemExit) as ec:
            self.run_command('--rev-range HEAD^1..HEAD {0}'.format(self.gitrepodir))

        self.assertSystemExitCode(ec.exception, 0)

        self.assertResults(u"""
            ▾  plugin01

            ⚠  line 1: c.txt
                c is +

            {0}  Jig ran 1 plugin
                Info 0 Warn 1 Stop 0
            """.format(ATTENTION), self.output)

    def test_defaults_latest_commit(self):
        """
        Without a revision range it uses HEAD^1..HEAD.
        """
        with self.assertRaises(SystemExit) as ec:
            self.run_command('{0}'.format(self.gitrepodir))

        self.assertResults(u"""
            ▾  plugin01

            ⚠  line 1: c.txt
                c is +

            {0}  Jig ran 1 plugin
                Info 0 Warn 1 Stop 0
            """.format(ATTENTION), self.output)

    def test_reports_two_commits(self):
        """
        With a range indicating two commits it reports on both.
        """
        with self.assertRaises(SystemExit) as ec:
            self.run_command('--rev-range HEAD~2..HEAD {0}'.format(self.gitrepodir))

        self.assertSystemExitCode(ec.exception, 0)

        self.assertResults(u"""
            ▾  plugin01

            ⚠  line 1: c.txt
                c is +

            ⚠  line 1: b.txt
                b is +

            {0}  Jig ran 1 plugin
                Info 0 Warn 2 Stop 0
            """.format(ATTENTION), self.output)

    def test_reports_only_one_plugin(self):
        """
        If a plugin is given it will only report on that plugin.
        """
        with self.assertRaises(SystemExit) as ec:
            self.run_command('--plugin plugin01 --rev-range HEAD^1..HEAD {0}'.format(self.gitrepodir))

        self.assertSystemExitCode(ec.exception, 0)

        self.assertResults(u"""
            ▾  plugin01

            ⚠  line 1: c.txt
                c is +

            {0}  Jig ran 1 plugin
                Info 0 Warn 1 Stop 0
            """.format(ATTENTION), self.output)



import jig
from jig.commands.base import BaseCommand

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Show Jig\'s version number',
    usage='jig version [-h]')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        with self.out() as out:
            out.append('{0}'.format(jig.__version__))



# coding=utf-8
from tempfile import mkdtemp

import jig
from jig.tests.testcase import CommandTestCase
from jig.commands import version


class TestVersionCommand(CommandTestCase):

    """
    Test the version subcommand.

    """
    command = version.Command

    def test_shows_version(self):
        """
        Shows Jig's current version.
        """
        self.run_command()

        self.assertResults(
            '{0}'.format(jig.__version__),
            self.output)



# coding=utf-8
import sys

from mock import Mock, patch

from jig.exc import PluginError
from jig.entrypoints import main
from jig.tests.testcase import JigTestCase, ViewTestCase, CommandTestCase
from jig.commands.base import (
    list_commands, create_view, add_plugin, BaseCommand)


class TestCommands(ViewTestCase):

    """
    Test the main parts of the command-line utility.

    """

    help_output_marker = '''
        usage: jig [-h] COMMAND

        optional arguments:
          -h, --help  show this help message and exit'''

    def setUp(self):
        self.view = create_view()

        self.view.collect_output = True
        self.view.exit_on_exception = False

    def test_main(self):
        """
        Will output help if no arguments are given
        """
        with patch.object(sys, 'stdout') as p:
            main()

        output = ''
        for stdout_call in p.write.call_args_list:
            output += stdout_call[0][0]

        self.assertResultsIn(self.help_output_marker, output)

    def test_main_help(self):
        """
        Will provide help menu when ran with no arguments.
        """
        commands = list_commands()

        self.view.print_help(commands)

        self.assertResultsIn(self.help_output_marker, self.output)


class TestBaseCommand(CommandTestCase):

    """
    Test our base command class.

    """
    def test_abstract_process(self):
        """
        The process method is abstract.
        """
        class MissingProcessCommand(BaseCommand):
            parser = Mock()

        with self.assertRaises(NotImplementedError):
            MissingProcessCommand([])


class MockUUID(object):
    hex = 'abcdef1234567890'


class TestAddPlugin(JigTestCase):

    """
    Add a plugin from url or local file system.

    """
    def setUp(self):
        super(TestAddPlugin, self).setUp()

        self.pm = Mock()

        self.clone_patch = patch('jig.commands.base.clone')
        self.clone = self.clone_patch.start()

        self.rmtree_patch = patch('jig.commands.base.rmtree')
        self.rmtree = self.rmtree_patch.start()

        self.uuid_patch = patch('jig.commands.base.uuid')
        self.uuid = self.uuid_patch.start()
        self.uuid.return_value = MockUUID

    def tearDown(self):
        super(TestAddPlugin, self).tearDown()

        self.clone_patch.stop()
        self.rmtree_patch.stop()
        self.uuid_patch.stop()

    def test_add_file_system(self):
        """
        Adds a file system plugin.
        """
        add_plugin(self.pm, '/a/b/c', self.gitrepodir)

        # Since this was from the file system clone was not called
        self.assertFalse(self.clone.called)

        # The plugin manager add() method was called with the location verbatim
        self.pm.add.assert_called_with('/a/b/c')

    def test_add_file_system_error_skips_cleanup(self):
        """
        Cleanup only occurs if the plugin is URL-based
        """
        self.pm.add.side_effect = PluginError

        # Make sure that the error does bubble out
        with self.assertRaises(PluginError):
            add_plugin(self.pm, '/a/b/c', self.gitrepodir)

        # Since we didn't clone a repository, we should not clean anything up
        self.assertFalse(self.rmtree.called)

    def test_url_without_branch(self):
        """
        URL without a branch specifier.
        """
        add_plugin(self.pm, 'http://a.b/c', self.gitrepodir)

        # Since this was a URL clone was called with the full URL
        self.clone.assert_called_with(
            'http://a.b/c',
            '{0}/.jig/plugins/{1}'.format(self.gitrepodir, MockUUID.hex),
            None)

        # The plugin manager was given the newly cloned location
        self.pm.add.assert_called_with(
            '{0}/.jig/plugins/{1}'.format(self.gitrepodir, MockUUID.hex))

        # Since things went well, the cleanup function was not ran
        self.assertFalse(self.rmtree.called)

    def test_url_with_branch(self):
        """
        URL with a branch specifier.
        """
        add_plugin(self.pm, 'http://a.b/c@branch', self.gitrepodir)

        # Since this was a URL clone was called with the full URL
        self.clone.assert_called_with(
            'http://a.b/c',
            '{0}/.jig/plugins/{1}'.format(self.gitrepodir, MockUUID.hex),
            'branch')

    def test_cleanup_on_error_with_url(self):
        """
        Cleanup occurs on a plugin error.
        """
        self.pm.add.side_effect = PluginError

        # Make sure that the error does bubble out
        with self.assertRaises(PluginError):
            add_plugin(self.pm, 'http://a.b/c', self.gitrepodir)

        self.rmtree.assert_called_with(
            '{0}/.jig/plugins/{1}'.format(self.gitrepodir, MockUUID.hex))



import json
import sys
from datetime import datetime

from git import Repo

from jig.exc import GitRepoNotInitialized
from jig.conf import PLUGIN_CHECK_FOR_UPDATES
from jig.gitutils.checks import repo_jiginitialized
from jig.gitutils.branches import parse_rev_range
from jig.diffconvert import GitDiffIndex
from jig.plugins import get_jigconfig, PluginManager
from jig.plugins.tools import (
    set_jigconfig, last_checked_for_updates, plugins_have_updates,
    set_checked_for_updates, update_plugins)
from jig.commands import get_command, list_commands
from jig.output import ConsoleView

try:
    from collections import OrderedDict
except ImportError:   # pragma: no cover
    from ordereddict import OrderedDict


def _diff_for(gitrepo, rev_range=None):
    """
    Get a list of :py:class:`git.diff.Diff` objects for the repository.

    :param git.repo.base.Repo gitrepo: Git repository
    :param string rev_range: optional revision to use instead of the Git index
    """
    if rev_range:
        commit_a, commit_b = parse_rev_range(gitrepo.working_dir, rev_range)

        return commit_a.diff(commit_b)
    else:
        # Assume we want a diff between what is staged and HEAD
        try:
            return gitrepo.head.commit.diff()
        except ValueError:
            return None


class Runner(object):

    """
    Runs jig in a Git repo.

    """
    def __init__(self, view=None):
        self.view = view or ConsoleView()

    def fromhook(self, gitrepo):
        """
        Main entry point called from pre-commit hook.

        :param unicode gitrepo: path to the Git repository
        """
        return self.main(gitrepo)

    def main(self, gitrepo, plugin=None, rev_range=None, interactive=True):
        """
        Run Jig on the given Git repository.

        :param unicode gitrepo: path to the Git repository
        :param unicode plugin: the name of the plugin to run, if None then run
            all plugins
        :param unicode rev_range: the revision range to use instead of the Git
            index
        :param bool interactive: if True then the user will be prompted to
            commit or cancel when any messages are generated by the plugins.
        """
        sys.stdin = open('/dev/tty')

        if interactive:
            # Check to see if the plugins need updating
            now = datetime.utcnow()

            with self.view.out():
                last_checked = last_checked_for_updates(gitrepo) or \
                    datetime.fromtimestamp(0)

            if now > last_checked + PLUGIN_CHECK_FOR_UPDATES:
                self.update_plugins(gitrepo)

        results = self.results(gitrepo, plugin=plugin, rev_range=rev_range)

        report_counts = self.view.print_results(results)

        if interactive and report_counts and sum(report_counts):
            # Git will run a pre-commit hook with stdin pointed at /dev/null.
            # We will reconnect to the tty so that raw_input works.
            while True:
                try:
                    answer = raw_input(
                        '\nCommit anyway (hit "c"), or stop (hit "s"): ')
                except KeyboardInterrupt:
                    sys.exit(1)
                if answer and answer[0].lower() == 's':
                    sys.exit(1)
                elif answer and answer[0].lower() == 'c':
                    break

        sys.exit(0)

    def fromconsole(self, argv):
        """
        Console entry point for the jig script.

        Where ``argv`` is ``sys.argv``.
        """
        # Quick copy
        argv = argv[:]
        # Our script is the first element
        argv.pop(0)

        try:
            # Next argument is the command
            command = get_command(argv.pop(0))
            command(argv)   # pragma: no cover
        except (ImportError, IndexError):
            # If it's empty
            self.view.print_help(list_commands())

    def update_plugins(self, gitrepo):
        """
        Prompt the user to update the plugins if available.

        :params string gitrepo: path to the Git repository
        """
        with self.view.out() as out:
            out.append(u'Checking for plugin updates\u2026')

        # Examine the remotes of the installed plugins
        if not plugins_have_updates(gitrepo):
            # If we don't have updates, change the date forward for the next
            # check to be a few days from now.
            set_jigconfig(gitrepo, set_checked_for_updates(gitrepo))
            # No updates, so nothing else we need to do
            return False

        # We have updates, ask the user if they want to fetch from the
        # remote and install them
        while True:
            try:
                answer = raw_input(
                    '\nPlugin updates are available, install ("y"/"n"): ')
            except KeyboardInterrupt:
                # If the user CTRL-C's out, leave the last date checked for
                # updates alone. Their intention with this is not really a
                # yes or a now so play it safe.
                return False
            else:
                # No KeyboardInterrupt, this is good enough to go ahead and
                # move the date we checked for updates to "now".
                set_jigconfig(gitrepo, set_checked_for_updates(gitrepo))
                # We now have a possible answer, do the appropriate thing
                if answer and answer[0].lower() == 'y':
                    update_plugins(gitrepo)
                    return True
                if answer and answer[0].lower() == 'n':
                    return False

    def results(self, gitrepo, plugin=None, rev_range=None):
        """
        Run jig in the repository and return results.

        Results will be a dictionary where the keys will be individual plugins
        and the value the result of calling their ``pre_commit()`` methods.

        :param unicode gitrepo: path to the Git repository
        :param unicode plugin: the name of the plugin to run, if None then run
            all plugins
        :param unicode rev_range: the revision range to use instead of the Git
            index
        """
        self.gitrepo = gitrepo

        # Is this repository initialized to use jig on?
        with self.view.out() as out:
            if not repo_jiginitialized(self.gitrepo):
                raise GitRepoNotInitialized(
                    'This repository has not been initialized.')

        pm = PluginManager(get_jigconfig(self.gitrepo))

        # Check to make sure we have some plugins to run
        with self.view.out() as out:
            if len(pm.plugins) == 0:
                out.append(
                    'There are no plugins installed, '
                    'use jig install to add some.')
                return

            self.repo = Repo(gitrepo)

            diff = _diff_for(self.repo, rev_range)

            if diff is None:
                # No diff on head, no commits have been written yet
                out.append(
                    'This repository is empty, jig needs at '
                    'least 1 commit to continue.')
                # Let execution continue so they *can* commit that first
                # changeset. This is a special mode that should not cause Jig
                # to exit with non-zero.
                return

            if len(diff) == 0:
                # There is nothing changed in this repository, no need for
                # jig to run so we exit with 0.
                out.append(
                    'No staged changes in the repository, skipping jig.')
                return

        # Our git diff index is an object that makes working with the diff much
        # easier in the context of our plugins.
        gdi = GitDiffIndex(self.gitrepo, diff)

        # Go through the plugins and gather up the results
        results = OrderedDict()
        for installed in pm.plugins:
            if plugin and installed.name != plugin:
                # This plugin doesn't match the requested
                continue

            retcode, stdout, stderr = installed.pre_commit(gdi)

            try:
                # Is it JSON data?
                data = json.loads(stdout)
            except ValueError:
                # Not JSON
                data = stdout

            results[installed] = (retcode, data, stderr)

        return results



from jig.exc import PluginError
from jig.plugins import (
    get_jigconfig, set_jigconfig, PluginManager)
from jig.plugins.tools import read_plugin_list
from jig.commands.base import BaseCommand, add_plugin
from jig.commands.hints import USE_RUNNOW

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Install a list of Jig plugins from a file',
    usage='jig install [-h] [-r GITREPO] [PLUGINSFILE]')

_parser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_parser.add_argument(
    'pluginsfile',
    help='Path to a file containing the location of plugins to install, '
    'each line of the file should contain URL|URL@BRANCH|PATH')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        path = argv.path
        plugins_file = argv.pluginsfile

        with self.out() as out:
            try:
                plugin_list = read_plugin_list(plugins_file)
            except IOError as e:
                # Grab the human-readable part of the IOError and raise that
                raise PluginError(e[1])

            for plugin in plugin_list:
                config = get_jigconfig(path)
                pm = PluginManager(config)

                try:
                    added = add_plugin(pm, plugin, path)
                except Exception as e:
                    out.append(
                        'From {0}:\n - {1}'.format(
                            plugin, e))
                    continue

                set_jigconfig(path, pm.config)

                out.append('From {0}:'.format(plugin))
                for p in added:
                    out.append(
                        ' - Added plugin {0} in bundle {1}'.format(
                            p.name, p.bundle))

            out.extend(USE_RUNNOW)



from jig.commands.base import BaseCommand
from jig.runner import Runner

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Run plugins on staged changes and show the results',
    usage='jig runnow [-h] [-p PLUGIN] [PATH]')

_parser.add_argument(
    '--plugin', '-p',
    help='Only run this specific named plugin')
_parser.add_argument(
    'path', nargs='?', default='.',
    help='Path to the Git repository')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        path = argv.path

        # Make the runner use our view
        runner = Runner(view=self.view)

        runner.main(path, plugin=argv.plugin, interactive=False)



import re
from unicodedata import normalize
from os import listdir, walk, makedirs, unlink, chdir, getcwd
from os.path import join, dirname, isdir
from tempfile import mkdtemp
from shutil import copy2
from contextlib import contextmanager

from git import Repo

_punct_re = re.compile(r'[\t !"#$%&\'()*\-/<=>?@\[\\\]^_`{|},.]+')


def slugify(text, delim=u'-'):
    """
    Generates a an ASCII-only slug.

    .. Author Armin Ronacher: http://flask.pocoo.org/snippets/5/
    """
    result = []
    for word in _punct_re.split(unicode(text).lower()):
        nword = normalize('NFKD', word).encode('ascii', 'ignore')
        if nword:
            # Filter out non-printable
            pword = ''.join([i for i in nword if ord(i) > 31])
            result.append(pword)
    return unicode(delim.join(filter(bool, result)))


def indent(payload, by=4, character=u' '):
    """
    Indents a sequence of strings with whitespace.

    By default it will indent by 4 spaces. Change the amount of indent with
    ``by`` and the character that is used with ``character``.

    Example:

        >>> print(indent(u'Jig', by=6, character=u'-'))
        ------Jig

    """
    return_first = False
    if isinstance(payload, (basestring)):
        payload = [payload]
        return_first = True

    indented = []
    for line in payload:
        indented.append(''.join([unicode(character)] * by) + unicode(line))

    if return_first:
        return indented[0]
    return indented


@contextmanager
def cwd_bounce(dir):
    """
    Temporarily changes to a directory and changes back in the end.

    Where ``dir`` is the directory you wish to change to. When the context
    manager exits it will change back to the original working directory.

    Context manager will yield the original working directory and make that
    available to the context manager's assignment target.
    """
    original_dir = getcwd()

    try:
        chdir(dir)

        yield original_dir
    finally:
        chdir(original_dir)


class NumberedDirectoriesToGit(object):

    """
    Converts a directory listing of snapshots to a Git repository.

    Given a directory where the layout is this::

        $ ls -1 testrepo
        .
        ..
        01
        02
        03

    This object can be initialized like this::

        >>> repo = NumberedDirectoriesToGit('testrepo')

    The resulting object will be a GitPython ``git.Repo`` object where each
    commit corresponds to the numbered directory. Commit 1 will be the contents
    in `01`. Commit 2 will be the contents of `02`, etc.

    The easiest way to think about this is in terms of using ``git checkout``.
    Each directory is a snapshot of the detached head for that commit.

    Supports directories, deleting files, and file modifications. Everything
    you'd expect in a testing utility for creating Git repo fixtures.

    """
    def __init__(self, numdir):
        if not isdir(numdir):
            raise ValueError('Not a directory: {0}.'.format(numdir))

        self.numdir = numdir
        self.target = mkdtemp()
        self._repo = None

    @property
    def repo(self):
        """
        Does the conversion and returns the ``git.Repo`` object.
        """
        if not self._repo:
            self._repo = Repo.init(self.target)

            for d in sorted(listdir(self.numdir)):
                self._commit(self._repo, join(self.numdir, d))

        return self._repo

    def diffs(self):
        """
        Get a list of diffs for all commits.
        """
        repo = self.repo

        diffs = []
        for commit in repo.iter_commits():
            try:
                diffs.append(commit.parents[0].diff(commit))
            except IndexError:
                pass

        # Make the oldest first
        diffs.reverse()

        return diffs

    def _commit(self, repo, d):
        """
        Creates a new commit in the repository

        Where ``repo`` is a ``git.Repo`` object to create the commit in. And
        ``d`` is the directory you want the new commit to look like once the
        commit has been made.
        """
        # List files currently in our repository
        paths_current = set(self._flattendirectory(
            repo.working_dir, strip=self.target))

        # What needs to be added
        paths_new = set(self._flattendirectory(
            d, strip=d))

        paths_to_add = paths_new - paths_current
        paths_to_del = paths_current - paths_new
        paths_same = paths_new.intersection(paths_current)

        # Add
        for path in paths_to_add.union(paths_same):
            # Create the directory if needed
            directory = join(self.target, dirname(path))
            if not isdir(directory):
                makedirs(directory)

            copy2(
                join(d, path),
                join(self.target, path))

            repo.index.add([path])

        # Delete
        for path in paths_to_del:
            repo.index.remove([join(self.target, path)])

            unlink(join(self.target, path))

        repo.index.commit('Commit from numbered directory {0}'.format(d))

    def _flattendirectory(self, d, strip=None):
        """
        Recursively generates a list of paths from a directory.

        If ``strip`` is not provided it defaults to ``d``.
        """
        for (dirpath, dirname, filenames) in walk(d):
            for fn in filenames:
                sdir = dirpath.replace('{0}'.format(strip), '').lstrip('/')
                if sdir.startswith('.git'):
                    continue
                yield join(sdir, fn)



from shutil import rmtree
from os.path import join
from contextlib import nested
from datetime import datetime, timedelta

from mock import patch

from jig.tests.testcase import (
    RunnerTestCase, PluginTestCase, result_with_hint)
from jig.conf import PLUGIN_CHECK_FOR_UPDATES
from jig.commands.hints import GIT_REPO_NOT_INITIALIZED
from jig.tests.mocks import MockPlugin
from jig.exc import ForcedExit
from jig.plugins import set_jigconfig, Plugin
from jig.runner import Runner


class TestRunner(RunnerTestCase, PluginTestCase):

    """
    Runner is used by the command line script to run jig.

    """
    def setUp(self):
        super(TestRunner, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

    def test_simple_integration(self):
        """
        Will run the hook.
        """
        with patch.object(self.runner, 'results'):
            plugin = MockPlugin()
            # Empty results
            self.runner.results.return_value = {
                plugin: (0, '', '')}

            with self.assertRaises(SystemExit) as ec:
                self.runner.fromhook(self.gitrepodir)  # pragma: no branch

            self.assertSystemExitCode(ec.exception, 0)

            self.assertEqual(
                u'\U0001f44c  Jig ran 1 plugin, nothing to report\n',
                self.output)

    def test_will_not_prompt_if_no_messages(self):
        """
        No prompt is presented if there are no messages to communicate.
        """
        with patch.object(Runner, 'results'):
            # No results came back from any plugin
            Runner.results.return_value = []

            with nested(
                patch('jig.runner.raw_input', create=True),
                patch('jig.runner.sys'),
                self.assertRaises(SystemExit)
            ) as (ri, r_sys, ec):
                r_sys.exit.side_effect = SystemExit
                self.runner.main(self.gitrepodir)

        # Make sure that the call to raw_input never happened
        self.assertFalse(ri.called)

        # And we exited with 0 because there is no reason to stop the commit
        r_sys.exit.assert_called_once_with(0)

    def test_will_prompt_user(self):
        """
        User sees a prompt if there are messages.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.raw_input', create=True),
            patch('jig.runner.sys'),
            self.assertRaises(SystemExit)
        ) as (ri, r_sys, ec):
            # Fake the raw_input call to return 's'
            r_sys.exit.side_effect = SystemExit
            ri.return_value = 's'

            self.runner.main(self.gitrepodir)

        # The user was prompted about committing or canceling
        ri.assert_called_once_with(
            '\nCommit anyway (hit "c"), or stop (hit "s"): ')
        # When they said cancel we exited with non-zero
        r_sys.exit.assert_called_once_with(1)

    def test_will_prompt_but_continue_anyway(self):
        """
        The user can choose to continue with the commit anyway.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.raw_input', create=True),
            patch('jig.runner.sys')
        ) as (ri, r_sys):
            # Fake the raw_input call to return 'c'
            ri.return_value = 'c'

            self.runner.main(self.gitrepodir)

        # The user was prompted about committing or canceling
        ri.assert_called_once_with(
            '\nCommit anyway (hit "c"), or stop (hit "s"): ')
        # When they said cancel we exited with non-zero
        r_sys.exit.assert_called_once_with(0)

    def test_will_continue_to_prompt_until_correctly_answered(self):
        """
        The user must answer 'c' or 's' and nothing else.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.raw_input', create=True),
            patch('jig.runner.sys')
        ) as (ri, r_sys):
            # Fake the raw_input call to return 'c' only after giving
            # two incorrect options.
            ri.side_effect = ['1', '2', 'c']

            self.runner.main(self.gitrepodir)

        # raw_input was called 3 times until it received a proper response
        self.assertEqual(3, ri.call_count)

    def test_will_abort_on_keyboard_interrupt(self):
        """
        The user can CTRL-C out of it and the commit is canceled.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.raw_input', create=True),
            patch('jig.runner.sys'),
            self.assertRaises(SystemExit)
        ) as (ri, r_sys, ec):
            # Fake the raw_input call to return 'c'
            ri.side_effect = KeyboardInterrupt
            r_sys.exit.side_effect = SystemExit

            self.runner.main(self.gitrepodir)

        # We exited with 1 to indicate the commit should abort
        r_sys.exit.assert_called_once_with(1)


class TestRunnerPluginUpdates(RunnerTestCase, PluginTestCase):

    """
    From the hook the plugins can be updated.

    """
    def setUp(self):
        super(TestRunnerPluginUpdates, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        targets = (
            'jig.runner.sys',
            'jig.runner.datetime',
            'jig.runner.plugins_have_updates',
            'jig.runner.set_jigconfig',
            'jig.runner.set_checked_for_updates',
            ('jig.runner.raw_input', {'create': True}),
            'jig.runner.update_plugins')

        self._patches = []
        for target in targets:
            if isinstance(target, tuple):
                patched = patch(target[0], **target[1])
                target = target[0]
            else:
                patched = patch(target)

            self._patches.append(patched)

            basename = target.split('.')[-1]
            setattr(self, basename, patched.start())

        # For all tests, make the current date in the future
        self.datetime.utcnow.return_value = datetime.utcnow() + \
            PLUGIN_CHECK_FOR_UPDATES + timedelta(days=1)

    def tearDown(self):
        for patched in self._patches:
            patched.stop()

    def test_non_interactive(self):
        """
        Doesn't check if interactive is False.
        """
        self.runner.main(self.gitrepodir, interactive=False)

    def test_never_been_checked(self):
        """
        For existing Jig installations, there will be no last checked value.
        """
        # There are no updates for the plugins
        self.plugins_have_updates.return_value = False

        with patch('jig.runner.last_checked_for_updates') as lcu:
            # If there is no value for the last time a repository was checked
            # it will return 0.
            lcu.return_value = 0

            self.runner.main(self.gitrepodir)

        # The check to see if the plugins have updates was called
        self.assertTrue(self.plugins_have_updates.called)

    def test_checks_for_updates(self):
        """
        Will check for updates if it has been a while.
        """
        # There are no updates for the plugins
        self.plugins_have_updates.return_value = False

        self.runner.main(self.gitrepodir)

        # The plugins were checked to see if there is an update
        self.assertTrue(self.plugins_have_updates.called)

        # Since they were checked, the date was also moved forward
        self.assertTrue(self.set_checked_for_updates.called)

        # Things exited normally
        self.sys.exit.assert_called_with(0)

    def test_prompts_user_to_update(self):
        """
        Will ask to install updates but the answer is no.
        """
        # This time there are updates to install
        self.plugins_have_updates.return_value = True

        # The answer will be "n"
        self.raw_input.side_effect = ['n']

        self.runner.main(self.gitrepodir)

        # They did give a valid answer, so the date was moved
        self.assertTrue(self.set_checked_for_updates.called)

        # The plugins were not updated though
        self.assertFalse(self.update_plugins.called)

    def test_prompts_until_they_answer_correctly(self):
        """
        Continues until a proper answer is given to the question.
        """
        self.plugins_have_updates.return_value = True

        # Answer a couple of times with junk, and then say no
        self.raw_input.side_effect = ['junk', 'foo', 'n']

        self.runner.main(self.gitrepodir)

        # The question was asked until a proper response was given
        self.assertEqual(3, self.raw_input.call_count)

    def test_keyboard_interrupt(self):
        """
        While being asked a question CTRL-C is pressed.
        """
        self.plugins_have_updates.return_value = True

        # Answer a couple of times with junk, and then say no
        self.raw_input.side_effect = KeyboardInterrupt

        self.runner.main(self.gitrepodir)

        # It exited just fine, no errors
        self.sys.exit.assert_called_with(0)

        # Since it was a CTRL-C, we shouldn't move the date forward
        self.assertFalse(self.set_checked_for_updates.called)

        # And the plugins were not updated
        self.assertFalse(self.update_plugins.called)

    def test_will_update_plugins(self):
        """
        If the answer is yes, the plugins are updated.
        """
        self.plugins_have_updates.return_value = True

        # The answer to update the plugins is yes
        self.raw_input.return_value = 'y'

        self.runner.main(self.gitrepodir)

        # Exited normally
        self.sys.exit.assert_called_with(0)

        # We moved the next date the plugins should be checked forward
        self.assertTrue(self.set_jigconfig.called)
        self.assertTrue(self.set_checked_for_updates.called)

        # And the plugins were updated
        self.assertTrue(self.update_plugins.called)


class TestRunnerResults(RunnerTestCase, PluginTestCase):

    """
    Can the runner iterate over the plugins and gather results.

    """
    def setUp(self):
        super(TestRunnerResults, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

    def test_uninitialized_repo(self):
        """
        The .jig directory has not been initialized.
        """
        # Remove the .jig directory, effectively un-initializing our repository
        rmtree(join(self.gitrepodir, '.jig'))

        with self.assertRaises(ForcedExit) as ec:
            self.runner.results(self.gitrepodir)

        self.assertEqual('1', str(ec.exception))
        self.assertResults(
            result_with_hint(
                u'This repository has not been initialized.',
                GIT_REPO_NOT_INITIALIZED),
            self.error)

    def test_no_plugins(self):
        """
        If there is a .jig directory without any plugins.
        """
        self.runner.results(self.gitrepodir)

        self.assertEqual(
            'There are no plugins installed, use jig '
            'install to add some.\n',
            self.output)

    def test_empty_repository(self):
        """
        If .jig is ran on a repository that hasn't had any commits
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.runner.results(self.gitrepodir)

        self.assertEqual(
            'This repository is empty, jig needs at '
            'least 1 commit to continue.\n',
            self.output)

    def test_no_diff(self):
        """
        If .jig is ran on a repository without any changes.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.commit(
            self.gitrepodir,
            name='a.txt',
            content='a')

        self.runner.results(self.gitrepodir)

        self.assertEqual(
            'No staged changes in the repository, skipping jig.\n',
            self.output)

    def test_unstaged_one_file(self):
        """
        Ran on a repository with an unstaged file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Add the first commit because we have to have it
        self.commit(
            self.gitrepodir,
            name='a.txt',
            content='a')

        # We've created this but not added it to the index
        self.create_file(
            self.gitrepodir,
            name='b.txt',
            content='b')

        self.runner.results(self.gitrepodir)

        self.assertEqual(
            'No staged changes in the repository, skipping jig.\n',
            self.output)

    def test_staged_one_file(self):
        """
        Ran on a repository with a staged file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Add the first commit because we have to have it
        self.commit(
            self.gitrepodir,
            name='a.txt',
            content='a')

        # Create a new file an stage it to the index
        self.stage(
            self.gitrepodir,
            name='b.txt',
            content='b')

        results = self.runner.results(self.gitrepodir)

        # One plugin ran, we should have results from it
        self.assertEqual(1, len(results))

        # It's plugin01
        self.assertEqual('plugin01', results.keys()[0].name)

        retcode, stdout, stderr = results.items()[0][1]

        # The return code is 0
        self.assertEqual(0, retcode)
        # We auto-convert to an object
        self.assertEqual({u'b.txt': [[1, u'warn', u'b is +']]}, stdout)
        # And no errors occurred here
        self.assertEqual('', stderr)

    def test_modified_one_file(self):
        """
        One modified and staged file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Add the first commit because we have to have it
        self.commit(
            self.gitrepodir,
            name='a.txt',
            content='a')

        # We've created this but not added it to the index
        self.stage(
            self.gitrepodir,
            name='a.txt',
            content='aaa')

        results = self.runner.results(self.gitrepodir)

        _, stdout, _ = results.items()[0][1]

        self.assertEqual(
            {u'a.txt': [[1, u'warn', u'a is -'],
                        [1, u'warn', u'aaa is +']]},
            stdout)

    def test_deleted_one_file(self):
        """
        Delete one file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.commit(
            self.gitrepodir,
            name='a.txt',
            content='a')

        # Now stage a file for removal
        self.stage_remove(self.gitrepodir, name='a.txt')

        results = self.runner.results(self.gitrepodir)

        _, stdout, _ = results.items()[0][1]

        # We should see it being removed
        self.assertEqual({u'a.txt': [[1, u'warn', u'a is -']]}, stdout)

    def test_specific_plugin(self):
        """
        Filter to results to a specific file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.commit(
            self.gitrepodir,
            name='a.txt',
            content='a')

        self.stage(
            self.gitrepodir,
            name='b.txt',
            content='b')

        # We can filter to the one that is already installed
        self.assertEqual(
            1,
            len(self.runner.results(self.gitrepodir, plugin='plugin01')))

        # If we try to filter on a non-existent plugin we get no results
        self.assertEqual(
            0,
            len(self.runner.results(self.gitrepodir, plugin='notinstalled')))

    def test_handles_non_json_stdout(self):
        """
        Supports non-JSON output from the plugin.
        """
        with patch.object(Plugin, 'pre_commit'):
            Plugin.pre_commit.return_value = (
                0, 'Test non-JSON output', '')

            self._add_plugin(self.jigconfig, 'plugin01')
            set_jigconfig(self.gitrepodir, config=self.jigconfig)

            self.commit(
                self.gitrepodir,
                name='a.txt',
                content='a')

            self.stage(
                self.gitrepodir,
                name='b.txt',
                content='b')

            results = self.runner.results(self.gitrepodir)

            _, stdout, _ = results.items()[0][1]

        # And we can still get the output even though it's not JSON
        self.assertEqual('Test non-JSON output', stdout)

    def test_handles_retcode_1_with_stderr(self):
        """
        Handles non-zero return codes and data written to stderr.
        """
        with patch.object(Plugin, 'pre_commit'):
            Plugin.pre_commit.return_value = (
                1, '', 'Something went horribly wrong')

            self._add_plugin(self.jigconfig, 'plugin01')
            set_jigconfig(self.gitrepodir, config=self.jigconfig)

            self.commit(
                self.gitrepodir,
                name='a.txt',
                content='a')

            self.stage(
                self.gitrepodir,
                name='b.txt',
                content='b')

            results = self.runner.results(self.gitrepodir)

            retcode, _, stderr = results.items()[0][1]

        self.assertEqual(1, retcode)
        self.assertEqual(
            'Something went horribly wrong',
            stderr)


class TestRunnerRevRange(RunnerTestCase, PluginTestCase):

    """
    Runner will take an options revision range instead of using staged files.

    """
    def setUp(self):
        super(TestRunnerRevRange, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        for letter in ['a', 'b', 'c']:
            self.commit(
                self.gitrepodir,
                name='{0}.txt'.format(letter),
                content=letter)

    def file_changes(self, results):
        """
        Get the file changes for the given results.
        """
        return results.items()[0][1][1]

    def test_bad_rev_range_format(self):
        """
        If a range is not formatted correctly.
        """
        with self.assertRaises(ForcedExit) as ec:
            results = self.runner.results(self.gitrepodir, rev_range="BAR:BAZ")

        self.assertEqual(
            'BAR:BAZ\n\nThe revision range is not in a valid format.\n\n'
            'Use "REV_A..REV_B" to specify the revisions that Jig should operate '
            'against.\n',
            self.error)

    def test_non_existent_rev_range(self):
        """
        If a range is given that does not exist.
        """
        with self.assertRaises(ForcedExit) as ec:
            results = self.runner.results(self.gitrepodir, rev_range="FOO..BAR")

        self.assertEqual(
            'FOO..BAR\n\nThe revision specified is formatted correctly but one '
            'of both of the revisions\ncould not be found.\n',
            self.error)

    def test_existing_rev_range(self):
        """
        Valid revision range returns results.
        """
        results = self.runner.results(self.gitrepodir, rev_range="HEAD^1..HEAD")

        self.assertEqual(1, len(self.file_changes(results)))

    def test_multiple_rev_range(self):
        """
        Valid revision that includes two changes returns both.
        """
        results = self.runner.results(self.gitrepodir, rev_range="HEAD~2..HEAD")

        self.assertEqual(2, len(self.file_changes(results)))



# coding=utf-8
from os.path import join
from tempfile import mkdtemp
from ConfigParser import SafeConfigParser

from jig.tests.testcase import (
    CommandTestCase, PluginTestCase,
    cd_gitrepo, result_with_hint)
from jig.commands import config
from jig.exc import ForcedExit
from jig.plugins import (
    set_jigconfig, get_jigconfig, create_plugin,
    PluginManager)
from jig.commands.hints import (
    NO_PLUGINS_INSTALLED, CHANGE_PLUGIN_SETTINGS, INVALID_CONFIG_KEY)


class TestPluginCommand(CommandTestCase, PluginTestCase):

    """
    Test the plugin command.

    """
    command = config.Command

    def setUp(self):
        super(TestPluginCommand, self).setUp()

        self.plugindir = mkdtemp()

    def _add_plugin(self, plugin_dir):
        """
        Adds a plugin to the jig initialized Git repository.
        """
        config = get_jigconfig(self.gitrepodir)
        pm = PluginManager(config)
        pm.add(plugin_dir)
        set_jigconfig(self.gitrepodir, pm.config)

    def _clear_settings(self, gitrepodir):
        """
        Remove all plugin specific settings.
        """
        config = get_jigconfig(self.gitrepodir)
        pm = PluginManager(config)

        for section in pm.config.sections():
            if not section.startswith('plugin'):
                continue
            for option, value in pm.config.items(section):
                if option == 'path':
                    continue
                pm.config.remove_option(section, option)

        set_jigconfig(self.gitrepodir, pm.config)

    def _set(self, gitrepodir, bundle_name, plugin_name, key, value):
        """
        Change a setting for a plugin and save the Jig config.
        """
        config = get_jigconfig(self.gitrepodir)
        pm = PluginManager(config)

        pm.config.set(
            'plugin:{0}:{1}'.format(bundle_name, plugin_name),
            key, value)

        set_jigconfig(self.gitrepodir, pm.config)

    def _plugin_about(self, plugin_dir, settings_about):
        """
        Create about/help messages for specific plugin settings.
        """
        config_filename = join(plugin_dir, 'config.cfg')
        config = SafeConfigParser()

        with open(config_filename) as fh:
            config.readfp(fh)

        for key, value in settings_about.items():
            config.set('help', key, value)

        with open(config_filename, 'w') as fh:
            config.write(fh)

    @cd_gitrepo
    def test_list_no_plugins(self):
        """
        No plugins are installed and the list is empty.
        """
        # Without the path argument, this should default to the current working
        # directory.
        self.run_command('list')

        self.assertResults(
            result_with_hint(
                u'No plugins installed.',
                NO_PLUGINS_INSTALLED),
            self.output)

    def test_list_settings(self):
        """
        List settings from freshly installed plugins.

        The default plugin config and the plugins.cfg will be in sync.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01', settings={
                'a': '1', 'b': '2', 'c': '3'}))
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test02', name='plugin02', settings={
                'a': '11', 'b': '22', 'c': '33'}))

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(
            result_with_hint(
                u'''
                test01.plugin01.a=1
                test01.plugin01.b=2
                test01.plugin01.c=3
                test02.plugin02.a=11
                test02.plugin02.b=22
                test02.plugin02.c=33
                ''',
                CHANGE_PLUGIN_SETTINGS),
            self.output)

    def test_list_plugins_no_overridden_config(self):
        """
        Lists settings if there are no overridden configs.

        The default plugin settings will not be in plugins.cfg.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01', settings={
                'a': '1', 'b': '2', 'c': '3'}))

        self._clear_settings(self.gitrepodir)

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(
            result_with_hint(
                u'''
                test01.plugin01.a=1
                test01.plugin01.b=2
                test01.plugin01.c=3
                ''',
                CHANGE_PLUGIN_SETTINGS),
            self.output)

    def test_list_plugins_one_overridden_config(self):
        """
        Lists settings if there is one overridden setting.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01', settings={
                'a': '1', 'b': '2', 'c': '3'}))

        self._clear_settings(self.gitrepodir)

        # Override the setting for this plugin, it should take the place of the
        # default which is '1'
        self._set(self.gitrepodir, 'test01', 'plugin01', 'a', 'one')

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(
            result_with_hint(
                u'''
                test01.plugin01.a=one
                test01.plugin01.b=2
                test01.plugin01.c=3
                ''',
                CHANGE_PLUGIN_SETTINGS),
            self.output)

    def test_list_missing_settings_on_plugin(self):
        """
        If the installed plugin is missing settings.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01'))

        plugin_config_filename = join(
            self.plugindir, 'plugin01', 'config.cfg')

        with open(plugin_config_filename, 'r') as fh:
            config = SafeConfigParser()
            config.readfp(fh)

        # What if this plugin has no settings whatsoever, perhaps the plugin
        # author just didn't have any and did not include that section
        config.remove_section('settings')

        with open(plugin_config_filename, 'w') as fh:
            config.write(fh)

        self.run_command('list -r {0}'.format(self.gitrepodir))

        self.assertResults(result_with_hint(
            u'''
            Installed plugins have no settings.
            ''',
            CHANGE_PLUGIN_SETTINGS),
            self.output)

    def test_config_set_plugin_not_found(self):
        """
        Setting cannot be changed because plugin is not installed.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01'))

        # plugin01 is installed but not the following
        with self.assertRaises(ForcedExit):
            self.run_command('set -r {0} notbundle.notplugin.a 111'.format(
                self.gitrepodir))

        self.assertResults(
            u'Could not locate plugin notplugin.',
            self.error)

    def test_config_invalid_key(self):
        """
        Invalid keys are handled.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01'))

        with self.assertRaises(ForcedExit):
            self.run_command('set -r {0} a:b:c 111'.format(self.gitrepodir))

        self.assertResults(result_with_hint(
            u'''
            a:b:c is an invalid config key.
            ''',
            INVALID_CONFIG_KEY),
            self.error)

    def test_config_set(self):
        """
        Setting can be changed.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01', settings={
                'a': '1', 'b': '2', 'c': '3'}))

        self.run_command(
            'set -r {0} test01.plugin01.a 111'.format(self.gitrepodir))

        # Setting was changed, no output to the console
        self.assertEqual('', self.output)

        # The setting was changed
        config = get_jigconfig(self.gitrepodir)

        self.assertEqual('111', config.get('plugin:test01:plugin01', 'a'))

    def test_config_no_plugins_installed(self):
        """
        If no plugins installed, about output mentions installing plugins.
        """
        self.run_command(
            'about -r {0}'.format(self.gitrepodir))

        self.assertResults(
            result_with_hint(
                u'No plugins installed.',
                NO_PLUGINS_INSTALLED),
            self.output)

    def test_config_no_help(self):
        """
        If the plugin has no settings, no about messages.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01'))

        self.run_command(
            'about -r {0}'.format(self.gitrepodir))

        self.assertResults(
            u'Installed plugins have no settings.',
            self.output)

    def test_config_no_help_but_defaults(self):
        """
        If the plugin has no help only the default is available.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01', settings={
                'a': '1', 'b': '2', 'c': '3'}))

        self.run_command(
            'about -r {0}'.format(self.gitrepodir))

        self.assertResults(
            u'''
            test01.plugin01.a
            (default: 1)

            test01.plugin01.b
            (default: 2)

            test01.plugin01.c
            (default: 3)
            ''',
            self.output)

    def test_config_help(self):
        """
        Help is available for each installed plugin setting that supports it.
        """
        self._add_plugin(create_plugin(
            self.plugindir, template='python',
            bundle='test01', name='plugin01', settings={
                'a': '1', 'b': '2', 'c': '3'}))

        # Make a about message for the ``a`` setting
        self._plugin_about(join(self.plugindir, 'plugin01'),
            {'a': 'turn a on'})

        self.run_command(
            'about -r {0}'.format(self.gitrepodir))

        self.assertResults(
            u'''
            test01.plugin01.a
            (default: 1)
               turn a on

            test01.plugin01.b
            (default: 2)

            test01.plugin01.c
            (default: 3)
            ''',
            self.output)



from urlparse import urlparse
from os import listdir
from os.path import join, dirname, isdir
from shutil import rmtree
from uuid import uuid4 as uuid

from jig.exc import PluginError
from jig.conf import JIG_DIR_NAME, JIG_PLUGIN_DIR
from jig.output import ConsoleView
from jig.gitutils.remote import clone

_commands_dir = dirname(__file__)


def list_commands():
    """
    List the commands available.
    """
    commands = []
    for f in listdir(_commands_dir):
        if isdir(join(_commands_dir, f)):
            continue
        if not f.endswith('.py'):
            continue
        try:
            commands.append(get_command(f.split('.')[0]))
        except (ImportError, AttributeError):
            continue
    return commands


def get_command(name):
    """
    Gets an instance of the named jig sub-command.

    For example::

        >>> get_command('init')
        <jig.commands.init.Command object at 0x10048fed0>
    """
    mod = __import__(
        'jig.commands.{0}'.format(name.lower()),
        globals(), locals(), ['Command'], 0)
    return mod.Command


def create_view():
    """
    Creates a view the command can use to output data.

    This method is separated from the :py:class:`BaseCommand` class to
    facilitate testing. By mocking out this method, commands can use views that
    have been configured to collect output instead of sending it to the
    terminal.
    """
    return ConsoleView()


def add_plugin(pm, plugin, gitdir):
    """
    Adds a plugin by filename or URL.

    Where ``pm`` is an instance of :py:class:`PluginManager` and ``plugin``
    is either the URL to a Git Jig plugin repository or the file name of a
    Jig plugin. The ``gitdir`` is the path to the Git repository which will
    be used to find the :file:`.jig/plugins` directory.
    """
    # If this looks like a URL we will clone it first
    url = urlparse(plugin)

    if url.scheme:
        # This is a URL, let's clone it first into .jig/plugins
        # directory.
        plugin_parts = plugin.rsplit('@', 1)

        branch = None
        try:
            branch = plugin_parts[1]
        except IndexError:
            pass

        to_dir = join(gitdir, JIG_DIR_NAME, JIG_PLUGIN_DIR, uuid().hex)
        clone(plugin_parts[0], to_dir, branch)
        plugin = to_dir

    try:
        return pm.add(plugin)
    except PluginError:
        # Clean-up the cloned directory becuase this wasn't installed correctly
        if url.scheme:
            rmtree(plugin)

        raise


def plugins_by_bundle(pm):
    """
    Organize plugins by bundle name.

    Returns a dict where the key is the bundle name and the value is a list
    of all plugins that are part of that bundle.
    """
    bundles = {}

    for plugin in pm.plugins:
        if plugin.bundle not in bundles:
            bundles[plugin.bundle] = []
        bundles[plugin.bundle].append(plugin)

    return bundles


def plugins_by_name(pm):
    """
    Organize plugins by plugin name.

    Returns a dict where the key is the plugin name and the value is a list
    of all plugins that have that name.
    """
    plugins = {}

    for plugin in pm.plugins:
        if plugin.name not in plugins:
            plugins[plugin.name] = []
        plugins[plugin.name].append(plugin)

    return plugins


class BaseCommand(object):

    """
    Base command for implementing script sub-commands.

    """
    def __init__(self, argv):
        """
        Parse the command line arguments and call process with the results.

        Where argv is a split string. See :py:module:`shlex`.
        """
        args = self.parser.parse_args(argv)

        # Setup something our command can use to send output
        self.view = create_view()
        # A shorter alias to the view's out decorator
        self.out = self.view.out

        # Finally, process the arguments
        self.process(args)

    def process(self, args):
        """
        Perform whatever operation this command is supposed to do.
        """
        raise NotImplementedError



import git
from git.exc import GitCommandError

from jig.exc import GitCloneError


def clone(repository, to_dir, branch=None):
    """
    Clone a Git repository to a directory.

    Where ``repository`` is a string representing a path or URL to the
    repository and ``to_dir`` is where the repository will be cloned.

    :param string repository: path or URL to the repository to clone
    :param string todir: where to clone the repository to
    :param string branch: branch to checkout instead of the repository's
        default
    """
    gitobj = git.Git()

    try:
        cmd = ['git', 'clone']

        if branch:
            cmd.extend(['--branch', branch])

        cmd.extend([repository, to_dir])

        gitobj.execute(cmd)

        return gitobj
    except git.GitCommandError as gce:
        raise GitCloneError(str(gce))


def remote_has_updates(repository):
    """
    Fetches the remote and check for available updates.

    :param string repository: path to the Git repository
    """
    try:
        repo = git.Repo(repository)

        # Get the latest tree from all remotes
        [i.fetch() for i in repo.remotes]

        active = repo.active_branch
        tracking = repo.active_branch.tracking_branch()

        is_different = active.commit != tracking.commit
        is_tracking_newer = \
            tracking.commit.committed_date > active.commit.committed_date
    except (AttributeError, GitCommandError, AssertionError):
        # The Python Git library issues some strange errors during
        # a fetch on occasion, so this "diaper"ish except is intended
        # to allow the process to continue without failing with a traceback.

        # Let the result be that new commits are available even though we had
        # an error. During a fetch this is typically the case.
        is_different = True
        is_tracking_newer = True

    return is_different and is_tracking_newer



import sys
import os
import imp
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.rst')).read()

version = imp.load_source(
    'jig', os.path.join(here, 'src', 'jig', '__init__.py')).__version__

install_requires = [
    'GitPython>=0.3.2RC1',
    'docutils>=0.9.1']

# Shims for missing stuff in Python 2.6
major, minor, patch, releaselevel, serial = sys.version_info
if major == 2 and minor < 7:
    install_requires += ['ordereddict==1.1', 'unittest2==0.5.1']

setup(
    name='jig',
    version=version,
    description="Check your code for stuff before you `git commit`",
    long_description=README + '\n\n' + NEWS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Version Control',
        'Topic :: Text Processing'
    ],
    keywords='git hooks code smell lint',
    author='Rob Madole',
    author_email='robmadole@gmail.com',
    url='http://github.com/robmadole/jig',
    license='MIT',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'jig = jig.entrypoints:main']}
)
